package org.ilrt.inkling.api;

import java.util.*;
import org.ilrt.inkling.query.*;
import org.ilrt.inkling.query.squish.*;

/*
Copyright University of Bristol, 2003

This software is released under the W3C license:
http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231
*/


/**
<pre>
A simple RDF Graph, inspired by http://www.w3.org/2001/12/rubyrdf/basicrdf.rb


The principal methods are

ask(), tell(), askSquish(); also askForObjects() and similar. Also
removeAll(), delete()
See GraphTest to see how they work.

 Graph gr = new Graph(null, uri, Util.RDFXML);
 gr.load();

 Vector classInstances 
=gr.askForSubjects(null,"http://www.w3.org/1999/02/22-rdf-syntax-ns#type",cl);

(returning a Vector of Nodes)

Vector superclasses
=gr.ask(cl,"http://www.w3.org/2000/01/rdf-schema#subClassOf",null);

(returning a Vector of Statements)

 java.sql.ResultSet r=gr.askSquish(query);

(returning a ResultSet)

gr.tell(new Statement(new Node(subject),new Node(predicate),new 
Node(object)));

gr.tellAll(vectorOfStatements v);

gr.delete(Statement s);

(s cannot have null Nodes in it as yet)

gr.removeAll();
</pre>
*/


public class Graph implements GraphInterface, SimpleQueryInterface {

    static Graph defaultdb = null;

//see base    String id = "";

    Hashtable fp;
    Hashtable bp;
    Hashtable ns;
    Vector statements;
    Vector statementIDs;
    Hashtable by_subject_id;
    Hashtable by_object_id;
    Hashtable by_predicate_id;
    Hashtable id_to_value_map;
    Hashtable value_to_id_map;
    Hashtable spo_index;
    String type=Util.RDFXML;
    String base=null;

    public Graph() {
        initialize(new Vector(), null);
    }

/**

base would normally be the source url

*/

    public Graph(String base) {
        initialize(null, base);
    }



    public Graph(Vector statements) {
        initialize(statements, null);
    }

/*
    public Graph(Vector statements, String base) {
        initialize(statements, base);
    }

*/
/**

Type should be taken from Util e.g. Util.RDFXML. You can also 
setType(Util.RDFXML)

Base is the source url

states is a list of Statements and can be null.

*/



    public Graph(String base,String type) {
        initialize(null, base);
	if(type!=null && (!type.equals(""))){
	setType(type);
//	System.out.println("!!!setting type "+type);
	}
    }

    public Graph(Vector states, String base,String type) {
        initialize(states, base);
	if(type!=null && (!type.equals(""))){
	setType(type);
//	System.out.println("!!!setting type "+type);
	}
    }

public void setType(String type){
this.type=type;
}
public String getType(){
return type;
}



/** 

loads in an RDF graph, using method depending on value of type (you can 
setType() or set it during initialization, or use gr.load(type)).

You need to load a Graph with content before using it, e.g.

 Graph gr = new Graph(null, uri, Util.RDFXML);
 gr.load();


*/

    public void load(){

	if(base!=null){
	System.out.println("Graph[1]: trying to load "+base+","+type);
	    load(type);
	}
    }


/** 

loads in an RDF graph, using method depending on value of type (you can 
setType() or set it during initialization, or use gr.load(type)).

You need to load a Graph with content before using it, e.g.

 Graph gr = new Graph(null, uri, null);
 gr.load(Util.RDFXML);


*/

    public void load(String type){

//find the correct class from Util
	String handler=Util.getHandler(type);

//check that have a name and not already loaded

	if(getID()!=null && (!getID().equals("")) 
&& handler!=null && (!handler.equals(""))){

		if((statements==null || statements.size()==0) ){

			try{

			LoaderInterface loader=(LoaderInterface)Class.forName(handler).newInstance();
			GraphInterface tmp=loader.loadGraph(getID(),(GraphInterface)this);

			}catch(Exception zz){
			System.err.println("Graph: type was "+type+" can't load "+zz);
			}
		}//end attempt to loadstatements

	}else{

	System.err.println("[a]nothing to load in graph - use setID first OR handler is null "+handler+" ,"+type);
	}

    }


/**

Creates a new storage environment, and tells all the statements in the
Vector to the graph

*/

    private void initialize(Vector states, String id) {

        if (id != null && (!id.equals(""))) {
            setID(id);
            //this.id=id;
        }


        fp = new Hashtable();
        bp = new Hashtable();
        ns = new Hashtable();

        by_subject_id = new Hashtable();
        by_object_id = new Hashtable();
        by_predicate_id = new Hashtable();
        id_to_value_map = new Hashtable();
        value_to_id_map = new Hashtable();
        spo_index = new Hashtable();

        statements = new Vector();
        statementIDs = new Vector();

        if (states != null) {

            for (Iterator iterator = states.listIterator();
                    iterator.hasNext();) {
                Statement statement = (Statement) iterator.next();
                //was hashtable
                try {
                    if (!statementIDs.contains(statement.getID())) {
                        statementIDs.addElement(statement.getID());
                        tell(statement);
                    }
                } catch (Exception rr) {
                    System.err.println("error "+rr);
                }
            }
        }

    }//end constructor


/**

base is usually the source url

*/

public void setBase(String base){
this.base=base;
}


/**

base is usually the source url

*/

public String getBase(){
return base;
}



/**

legacy - calls getBase()

*/

    public String getID() {
        return getBase();
    }

/**

legacy - calls setBase()

*/

    public void setID(String base) {
	this.setBase(base);
    }


/**

returns the full Vector of Statements from the Graph

*/
    public Vector getStatements() {
        return statements;
    }


/**

For Graphs without a uri (base), creates one with a default uri

*/

    public static GraphInterface getDefault() {

        if (defaultdb == null) {
            //System.out.print ("Getting default graph\n");
            defaultdb = new Graph(new Vector());
        }
        defaultdb.setID("http://rdfweb.org/graphs/default");
        return defaultdb;

    }


/**

Tell a Vector of Statements to the Graph

*/

    public void tellAll(Vector states) {


        for (Iterator iterator = states.listIterator();
                iterator.hasNext();) {
            Statement statement = (Statement) iterator.next();
            tell(statement);
        }

    }

/**

Tell one Statement to the Graph

*/

    public void tell(Statement statement) {

        //first index - just a list of statements
        //uniquesness?

        if (!statements.contains(statement)) {
            statements.addElement(statement);
            //	System.out.println("storing statement "+statements.size());
        }
        //new index - store statement under spo index
        //not sure if we need vectors or not.

        ///second index. changes ids, so needs to be high up
        ///changed to one to one
        //value to id map - one to one won't do...
        //s

        String sub_ids = (String) value_to_id_map.get(
                statement.getSubject().toString());

        if (sub_ids == null) {
            value_to_id_map.put(statement.getSubject().toString(),
                    statement.getSubject().getID());
            //System.out.println("NOT CHANGING o "+sub_ids);
        } else {
            statement.getSubject().setID(sub_ids);
            //System.out.println("CHANGING o "+statement.getSubject().getID());
        }

        //p
        String pred_ids = (String) value_to_id_map.get(
                statement.getPredicate().toString());

        if (pred_ids == null) {
            value_to_id_map.put(statement.getPredicate().toString(),
                    statement.getPredicate().getID());
            //System.out.println("NOT CHANGING p "+statement.getPredicate().getID());
        } else {
            statement.getPredicate().setID(pred_ids);
            //System.out.println("CHANGING p "+pred_ids);
        }

        ///o
        String obj_ids = (String) value_to_id_map.get(
                statement.getObject().toString());

        if (obj_ids == null) {
            value_to_id_map.put(statement.getObject().toString(),
                    statement.getObject().getID());
            //System.out.println("NOT CHANGING o "+statement.getObject().getID());
        } else {
            statement.getObject().setID(obj_ids);
            //System.out.println("CHANGING o "+obj_ids);
        }


        ///end second


        String key = statement.getSubject().getID() + " | "+
                statement.getPredicate().getID() + " | "+
                statement.getObject().getID();

        Vector spo = (Vector) spo_index.get(key);

        if (spo != null) {
            spo.addElement(statement);
        } else {
            spo = new Vector();
            spo.addElement(statement);
            spo_index.put(key, spo);
        }

        //store objects under subject+predicate ** libby changed to IDs **
        Vector sp_list =
                (Vector) fp.get(statement.getSubject().getID() + " | "+
                statement.getPredicate().getID());

        if (sp_list != null) {
//          System.out.print("Storing object under EXISTINGs/p,statement="+
//                    statement.toString());
//          System.out.print("IDS are "+statement.getSubject().getID()+" | "+
//                    statement.getPredicate().getID());
            //	sp_list.addElement(statement.getObject().getID());
            sp_list.addElement(statement);
        } else {
//            System.out.print("Storing object under NEW s/p,statement="+
 //                   statement.toString());
//            System.out.print("IDS are "+statement.getSubject().getID()+" | "+
//                    statement.getPredicate().getID());
            sp_list = new Vector();
            //	sp_list.addElement(statement.getObject().getID());
            sp_list.addElement(statement);
            fp.put(statement.getSubject().getID() + " | "+
                    statement.getPredicate().getID(), sp_list);
        }


        //store subjects under predicate+object ** libby changed to IDs **
        Vector po_list =
                (Vector) bp.get(statement.getPredicate().getID() + " | "+
                statement.getObject().getID());

        if (po_list != null) {
//            System.out.print("Storing subject under EXISTINGp/o,statement="+
//                    statement.toString());
            //System.out.println("OLD bp: IDS are "+statement.getPredicate().getID()+" | "+statement.getObject().getID());
            //	po_list.addElement(statement.getSubject().getID());
            po_list.addElement(statement);
        } else {
//            System.out.print("Storing object under NEW p/o,statement="+
//                    statement.toString());
            //	System.out.println("NEW bp: IDS are "+statement.getPredicate().getID()+" | "+statement.getObject().getID());
            po_list = new Vector();
            //	po_list.addElement(statement.getSubject().getID());
            po_list.addElement(statement);
            bp.put(statement.getPredicate().getID() + " | "+
                    statement.getObject().getID(), po_list);
        }

        //store s,p,o under id_to_value_map and value_to_id_map - need both?

        //id to value map
        //s
        Vector sub_values = (Vector) id_to_value_map.get(
                statement.getSubject().getID());
        if (sub_values == null) {
            /*
             	sub_values=new Vector();
             	sub_values.addElement(statement.getSubject().toString());
             	id_to_value_map.put(statement.getSubject().getID(), sub_values);
             */
            //objects better I think...
            sub_values = new Vector();
            sub_values.addElement(statement.getSubject());
            id_to_value_map.put(statement.getSubject().getID(), sub_values);
        } else {
            if (!sub_values.contains(statement.getSubject())) {
                sub_values.addElement(statement.getSubject());
            }
        }

        //p
        Vector pred_values = (Vector) id_to_value_map.get(
                statement.getPredicate().getID());
        if (pred_values == null) {
            pred_values = new Vector();
            pred_values.addElement(statement.getPredicate());
            id_to_value_map.put(statement.getPredicate().getID(),
                    pred_values);
        } else {
            if (!sub_values.contains(statement.getPredicate())) {
                pred_values.addElement(statement.getPredicate());
            }
        }


        //o
        Vector obj_values = (Vector) id_to_value_map.get(
                statement.getObject().getID());
        if (obj_values == null) {
            obj_values = new Vector();
            obj_values.addElement(statement.getObject());
            id_to_value_map.put(statement.getObject().getID(), obj_values);
        } else {
            if (!obj_values.contains(statement.getObject())) {
                obj_values.addElement(statement.getObject());
            }
        }



        //finally id->statement idexes for when we just have one value

        //subject
        Vector sub_map = (Vector) by_subject_id.get(
                statement.getSubject().getID());
        if (sub_map == null) {
            sub_map = new Vector();
            sub_map.addElement(statement);
            by_subject_id.put(statement.getSubject().getID(), sub_map);
        } else {
            if (!sub_map.contains(statement)) {
                sub_map.addElement(statement);
            }
        }

        //predicate
        Vector pred_map = (Vector) by_predicate_id.get(
                statement.getPredicate().getID());
        if (pred_map == null) {
            pred_map = new Vector();
            pred_map.addElement(statement);
            by_predicate_id.put(statement.getPredicate().getID(), pred_map);
        } else {
            if (!pred_map.contains(statement)) {
                pred_map.addElement(statement);
            }
        }

        //object
        Vector obj_map =
                (Vector) by_object_id.get(statement.getObject().getID());

        if (obj_map == null) {
            obj_map = new Vector();
            obj_map.addElement(statement);
            by_object_id.put(statement.getObject().getID(), obj_map);
        } else {
            if (!obj_map.contains(statement)) {
                obj_map.addElement(statement);
            }
        }


        ///end of indexing (phew!)

    }



/**

Handy method returning a Vector of object Nodes answering the query

*/

    public Vector askForObjects(Object ss, Object pp, Object oo) {

	Vector objects= new Vector();
	Vector statements=ask(ss,pp,oo);

	for (Iterator iterator = statements.listIterator();
	iterator.hasNext();) {
	Statement statement = (Statement) iterator.next();
	Node object=statement.getObject();
	objects.addElement(object);
	}
     return objects;

    }



/**

Handy method returning a Vector of subject Nodes answering the query

*/

    public Vector askForSubjects(Object ss, Object pp, Object oo) {

	Vector subjects= new Vector();
	Vector statements=ask(ss,pp,oo);

	for (Iterator iterator = statements.listIterator();
	iterator.hasNext();) {
	Statement statement = (Statement) iterator.next();
	Node subject=statement.getSubject();
	subjects.addElement(subject);
	}
     return subjects;

    }


    /**

    The simple triple query method - with Nodes
    GraphInterface method

     */


    public Vector askQuery(Node ss, Node pp, Node oo) {
        Query query = new Query(ss, pp, oo);
        return ask(query).getStatements();
    }


    /**

     Simple query method - with lazy strings

     */

    public Vector ask(Object ss, Object pp, Object oo) {

        Node s=null;
        Node p=null;
        Node o=null;

                if (ss instanceof java.lang.String){
                s=new Node((String)ss);
                }else if(ss instanceof org.ilrt.inkling.api.Node || ss==null){
                s=(Node)ss;
                }     
                else{
                System.err.println("Graph ask must be passed Strings or Nodes");
                return null;
                }

                if (pp instanceof java.lang.String){
                p=new Node((String)pp);
                }else if(pp instanceof org.ilrt.inkling.api.Node || pp==null){
                p=(Node)pp;//assume Node
                }
                else{
                System.err.println("Graph ask must be passed Strings or Nodes");
                return null;
                }
                if (oo instanceof java.lang.String){
                o=new Node((String)oo);
                }else if(oo instanceof org.ilrt.inkling.api.Node || oo==null){
                o=(Node)oo;//assume Node
                }     
                else{
                System.err.println("Graph ask must be passed Strings or Nodes");
                return null;
                }

        Query query = new Query(s, p, o);


        return ask(query).getStatements();
    }

/**

Ask with a squish string; returns a ResultSet object, not a Graph.

*/

    public java.sql.ResultSet askSquish(String query) {
        ComplexQuery cq=(ComplexQuery)SquishParser.parse(query);
        cq.getGraphs().addElement(this);
        QE qe=new QE();
        return qe.executeQuery(cq);

    }


/**

Ask with a query object

*/

    public Graph ask(Query query) {


        //ooo - dump all statements in the graph
        if (query.getPredicate() == null &&
                query.getSubject() == null && query.getObject() == null) {

            ///not sure about this
            //try just dumping the statements out

            return new Graph(statements);

        }



        //xox - dump all statements win the graph with a fixed predicate
        if (query.getPredicate() != null &&
                query.getSubject() == null && query.getObject() == null) {

            //first find the predicate's id(s) in the reverse lookup table
            String id = (String) value_to_id_map.get(
                    query.getPredicate().toString());

		Graph graph=new Graph();
		if(id!=null){	

	            Vector v = (Vector) by_predicate_id.get(id);
	            graph = new Graph(new Vector());
	            if (v != null) {
	                graph.tellAll(v);
	                //System.out.println("SIZE:[3] "+v.size());
	            }
		}
            return graph;

        }


        //oxx - dump all statements win the graph with a fixed subject
        if (query.getPredicate() == null &&
                query.getSubject() != null && query.getObject() == null) {

            //first find the subject's id(s) in the reverse lookup table

            //System.out.println("sub "+query.getSubject().toString());
            String id = (String) value_to_id_map.get(
                    query.getSubject().toString());
            Graph graph = new Graph(new Vector());
            //System.out.println("ID lookup table "+value_to_id_map.toString());

            Vector v = (Vector) by_subject_id.get(id);
            //System.out.println("v "+v);

            if (v != null) {
                graph.tellAll(v);
            }
            //System.out.println("GRAPH size oxx "+graph.getStatements().size());
            return graph;
        }


        //xxo - dump all statements win the graph with a fixed object
        if (query.getPredicate() == null &&
                query.getSubject() == null && query.getObject() != null) {

            //first find the object's id in the reverse lookup table

            String id = (String) value_to_id_map.get(
                    query.getObject().toString());
            Graph graph = new Graph(new Vector());
            Vector v = (Vector) by_object_id.get(id);
            if (v != null) {
                graph.tellAll(v);
            }
            return graph;
        }


        //xxx: is this statement in the graph?

        if (query.getPredicate() != null &&
                query.getSubject() != null && query.getObject() != null) {

            String sid = (String) value_to_id_map.get(
                    query.getSubject().toString());
            String pid = (String) value_to_id_map.get(
                    query.getPredicate().toString());
            String oid = (String) value_to_id_map.get(
                    query.getObject().toString());

            Graph graph = new Graph(new Vector());


            //a new index

            String key = sid + " | "+pid + " | "+oid;
            Vector spo = (Vector) spo_index.get(key);

            //System.out.println("SPO "+spo);

            if (spo != null) {
                graph.tellAll(spo);
            }

            return graph;
        }//end xxx


        //xxo - got s and p
        if (query.getPredicate() != null &&
                query.getSubject() != null && query.getObject() == null) {

            String sid = (String) value_to_id_map.get(
                    query.getSubject().toString());
            String pid = (String) value_to_id_map.get(
                    query.getPredicate().toString());

            Graph graph = new Graph(new Vector());

            Vector sp_result = (Vector) fp.get(sid + " | "+pid);
            //now statements
            if (sp_result != null && sp_result.size() != 0) {

                for (Iterator iterator = sp_result.listIterator();
                        iterator.hasNext();) {
                    //not sure....could be that we need id_to_val_map
                    Statement statement = (Statement) iterator.next();
                    graph.tell(statement);
                }

            }//end if

            return graph;
        }//end xxo


        //oxx - got o and p

        if (query.getPredicate() != null &&
                query.getSubject() == null && query.getObject() != null) {

            String oid = (String) value_to_id_map.get(
                    query.getObject().toString());
            String pid = (String) value_to_id_map.get(
                    query.getPredicate().toString());

            Graph graph = new Graph(new Vector());
            int count = 0;

            Vector op_result = (Vector) bp.get(pid + " | "+oid);
            if (op_result != null && op_result.size() != 0) {

                //System.out.println("opresult "+op_result.toString());
                for (Iterator iterator = op_result.listIterator();
                        iterator.hasNext();) {
                    Statement statement = (Statement) iterator.next();
                    graph.tell(statement);
                    //System.out.println("TELLING "+statement.toString());
                    count++;
                }

            }

            //	System.out.println("SIZE: "+graph.getStatements().size());
            //	System.out.println("SIZE:[2] "+count);
            return graph;

        }//end xxo



        ///xox - got s and o

        if (query.getPredicate() == null &&
                query.getSubject() != null && query.getObject() != null) {

            //hack for now - runs through dubjects and checks for the object in the statement.
            //needs a new index?

            //first find the subject's id in the reverse lookup table
            String sid = (String) value_to_id_map.get(
                    query.getSubject().toString());
            Graph graph = new Graph(new Vector());

            //get the statements in by_subject_id
            Vector stmnts = (Vector) by_subject_id.get(sid);

            for (Iterator iterator = stmnts.listIterator();
                    iterator.hasNext();) {
                Statement statement = (Statement) iterator.next();
                if (statement.getObject().toString().equals(
                        query.getObject().toString())) {
                    graph.tell(statement);
                    //System.out.println("TELLING "+statement.toString());
                }
            }
            //	System.out.println("GRAPH size xox "+graph.getStatements().size());
            return graph;
        }

        return new Graph(new Vector());

    }//end ask method



/**

Returns a string containing the contents of the Graph - for debugging

*/

    public String inspect() {


        StringBuffer sb = new StringBuffer();
        int numStatements = 0;

        Enumeration e = spo_index.keys();
        while (e.hasMoreElements()) {
            String key = (String) e.nextElement();
            Vector v = (Vector) spo_index.get(key);
            for (Iterator iterator = v.listIterator();
                    iterator.hasNext();) {
                Statement statement = (Statement) iterator.next();
                sb.append(statement.toString() + "\n");
                numStatements++;
            }

        }

        return "num statements: "+numStatements + "\n\n"+sb.toString();

    }

    /**

Shortcut for getting all object Nodes in the Graph

     */
    public Vector getObjects() {

        Vector v = new Vector();

        for (Iterator iterator = statements.listIterator();
                iterator.hasNext();) {
            Statement statement = (Statement) iterator.next();
            v.addElement(statement.getObject());

        }

        return v;
    }

/**

Shortcut for getting all subject Nodes in the Graph

*/
    public Vector getSubjects() {

        Vector v = new Vector();

        for (Iterator iterator = statements.listIterator();
                iterator.hasNext();) {
            Statement statement = (Statement) iterator.next();
            v.addElement(statement.getSubject());

        }

        return v;
    }

/**

Shortcut for getting all predicates (properties) in the Graph

*/

    public Vector getPredicates() {

        Vector v = new Vector();

        for (Iterator iterator = statements.listIterator();
                iterator.hasNext();) {
            Statement statement = (Statement) iterator.next();
            v.addElement(statement.getPredicate());
        }

        return v;
    }


/**

Remove all Statements in the Graph

*/

    public boolean removeAll(){
    statements=new Vector();
    spo_index=new Hashtable();
    fp=new Hashtable();
    bp=new Hashtable();
    by_subject_id=new Hashtable();
    by_predicate_id=new Hashtable();
    by_object_id=new Hashtable();
    return true;
    }


/**

Delete a specified Statement
@@fixme - what if blanks

*/

    public boolean delete(Statement statement) {


        //first index - just a list of statements
        //uniquesness?
        if (statements.contains(statement)) {
            statements.remove(statement);//will this work?
            //System.out.println("removing[1] "+statement);
        } else {
            System.out.println("could not delete");
        }

        ///second index - value to id map can be ignored
        ///third index

        //reset the ids

        String sid = (String) value_to_id_map.get(
                statement.getSubject().toString());
        String pid = (String) value_to_id_map.get(
                statement.getPredicate().toString());
        String oid = (String) value_to_id_map.get(
                statement.getObject().toString());


        //String key=statement.getSubject().getID()+" |
        //"+statement.getPredicate().getID()+" | "+statement.getObject().getID();

        String key = sid + " | "+pid + " | "+oid;

//        System.out.println("DELETING[2] "+key);

        Vector spo = (Vector) spo_index.get(key);

        if (spo != null) {
            spo.remove(statement);
        } else {
            System.out.println("can't remove from index1");
        }

        //remove objects under subject+predicate ** libby changed to IDs **
        //Vector sp_list=(Vector)fp.get(statement.getSubject().getID()+" |
        //"+statement.getPredicate().getID());

        Vector sp_list = (Vector) fp.get(sid + " | "+pid);

        if (sp_list != null) {
            //	print("removing object under EXISTING s/p, statement="+statement.toString());
            //	print("IDS are "+statement.getSubject().getID()+" | "+statement.getPredicate().getID());
            //System.out.println("removing[3] "+statement);

            sp_list.remove(statement);
        } else {
            System.out.println("can't remove from index2");
        }


        //remove subjects under predicate+object ** libby changed to IDs **
        //Vector po_list=(Vector)bp.get(statement.getPredicate().getID()+" |
        //"+statement.getObject().getID());

        Vector po_list = (Vector) bp.get(pid + " | "+oid);

        if (po_list != null) {
            po_list.remove(statement);
        } else {
            System.out.println("can't remove from index3");
        }

        //remove s,p,o under id_to_value_map and value_to_id_map - need both?
        //shouldnt need to cxhange this


        //finally id->statement idexes for when we just have one value
        //shouldnt need to chanfge this

	//fix this:

	//subject
        Vector sub_map = (Vector) by_subject_id.get(
                statement.getSubject().getID());
        if (sub_map != null) {
                sub_map.remove(statement);
        }

        //predicate
        Vector pred_map = (Vector) by_predicate_id.get(
                statement.getPredicate().getID());
        if (pred_map != null) {
            pred_map.remove(statement);
        } 

        //object
        Vector obj_map =
                (Vector) by_object_id.get(statement.getObject().getID());

        if (obj_map != null) {
            obj_map.remove(statement);
        } 

        return true;
    }

/**

Returns the number of unique Statements in the Graph

*/


public int getSize(){
  return statements.size();
}

/**

Returns the number of unique Statements in the Graph

*/


public int size(){
  return getSize();
}

/**

Returns all subjects and properties

*/

public Vector getNodes(){

        Vector v = new Vector();

        for (Iterator iterator = statements.listIterator();
                iterator.hasNext();) {
            Statement statement = (Statement) iterator.next();
            v.addElement(statement.getSubject());
            v.addElement(statement.getObject());
        }

        return v;

}

/**

Returns all properties

*/

public Vector getEdges(){
return getPredicates();
}


}//end class


