package org.ilrt.inkling.api;


import java.io.*;
import java.util.*;

/*
Copyright University of Bristol, 2003

This software is released under the W3C license:
http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231
*/

/**
<pre>
This is a very very basic Ntriples parser with no datatyping or language 
support.

It will try and read Ntriples document in and interpret it as a query or
an RDF file depending on which method is called.

Because it's here, theoretically you can use Tinkling on its own without 
any external classes; however you would probably be better off using RIO 
or ARP.

The N3Reader.java class uses Jena's N3 parser and code by Andy Seaborne, 
which returns Ntriples; therefore this class is also used there.
</pre>
*/


public class NTReader implements LoaderInterface{

//this is a ComplexQuery because this class can read queries expressed 
//as Ntriples.
GraphInterface g=new ComplexQuery();

public NTReader(){}

/**

LoaderInterface function

*/

    public GraphInterface loadGraph(String fn)  {

        g = new ComplexQuery();
        return loadGraph(fn,g);

    }

/**

LoaderInterface function

*/

    public GraphInterface loadGraph(String url, GraphInterface rdf){

            g = null;
            try {
                g = (GraphInterface) rdf;
                g.setBase(url);
		convertToStatement(url);

           } catch (Exception e1) {
                System.err.println(
                        "Download :something wrong NTriples[3]" + e1);
                e1.printStackTrace(); 
            }
            return g;

	}




/** 

Read in an Ntriples text file and return a ComplexQuery 
for more details see the RDF Query test work: 
http://www.w3.org/2003/03/rdfqr-tests/

An ntriples file can (unofficially and experimentally) be interpreted as a 
query, with bnodes representing variables. Predicates can also be 
blank/varianbles under this interpretation.

*/

    public ComplexQuery parseToQuery(String fn) {

        
        if (fn.startsWith("file:")) {      
            fn = fn.substring(5);            
        }
        

	try{
        RandomAccessFile in = new RandomAccessFile(fn, "r");
        String text = "";
        String line = "";
        while (line != null){   
	    if(!line.startsWith("#") && (!line.equals(""))){
		    convertToQuery(line);
	    }
            line = in.readLine();
	    
        }
	}catch(Exception e){
	System.err.println("NTParser error[2] "+e);
	}
       return (ComplexQuery)g;
     }



/**

Converts a String to an Ntriples query, where the string contains 
multiple ntriples lines. 

*/

    public ComplexQuery parseStringToQuery(String string)  {
	try{

        StringTokenizer st = new StringTokenizer(string, "\n");
        boolean doneFirst = false;
        while (st.hasMoreTokens()) {
            String sss = st.nextToken();
	    convertToQuery(sss);
            doneFirst = true;  
        }
        
	}catch(Exception e){
	System.err.println("NTParser error[1] "+e);
	}
	return (ComplexQuery)g;
     }



/**

Tries to parse a String as a single triple Query; adds it to the
GraphInterface g.

*/

private ComplexQuery convertToQuery(String in){

		if(!in.trim().equals("")){

		try{
		Query q=processTripleToQuery(in);
//System.out.println("Q "+q);


		   if(q!=null){
			((ComplexQuery)g).getTriples().addElement(q);
		   }

		}catch(Exception e){
		e.printStackTrace();
		}
		}

return (ComplexQuery)g;
}



/**

Nasty Ntriples parser, returning a Query given a String (assumed to be a
single line of NTriples)

*/

public Query processTripleToQuery(String part){

	Query trip = null;

	if(part!=null && (!part.trim().equals(""))){
	trip=new Query();
	int k=0;

		while(k<3){
		int i=part.indexOf(" ");
		if(k==2){
		i=part.lastIndexOf(".");
		}

	String node=part.substring(0,i).trim();

	part=part.substring(i+1).trim();

		if(node.startsWith("<")){
		Node r = Node.getResource(node.substring(1,node.lastIndexOf(">")));

			if(k==0){
			trip.setSubject(r); trip.setSubjectVariable(r.toString());
			}
			if(k==1){
			trip.setPredicate(r);trip.setPredicateVariable(r.toString());
			}
			if(k==2){
			trip.setObject(r);trip.setObjectVariable(r.toString());
			}
		}else 
		if(node.startsWith("_:")){
		String var=node.substring(2);


			if(!((ComplexQuery)g).getVariablesWeWant().contains(var)){

			((ComplexQuery)g).getVariablesWeWant().addElement(var);
			}


			if(!((ComplexQuery)g).getVariables().contains(var)){

			((ComplexQuery)g).getVariables().addElement(var);
			}


			if(k==0){
			trip.setSubjectVariable(var);
			}
			if(k==1){
			trip.setPredicateVariable(var);
			}
			if(k==2){
			trip.setObjectVariable(var);
			}
		}else 
		if(node.startsWith("\"")){
		Node r = Node.getLiteral(node.substring(1,node.lastIndexOf("\"")));
			if(k==0){System.out.println("error[1]");}
			if(k==1){System.out.println("error[2]");}
			if(k==2){trip.setObject(r);trip.setObjectVariable(r.toString());}
		}
		else if(node.startsWith("'")){
		Node r = Node.getLiteral(node.substring(1,node.lastIndexOf("'")));
			if(k==0){System.out.println("error[1]");}
			if(k==1){System.out.println("error[2]");}
			if(k==2){trip.setObject(r);trip.setObjectVariable(r.toString());}
		}

	k++;

	}

	}

return trip;
}


/** 

Read in an Ntriples text file by filename and return a Graph 

*/

    public Graph parseToStatement(String fn)  {
	GraphInterface g = new ComplexQuery();
	return parseToStatement(fn,g);
	}

/** 

read in an Ntriples text file by filename and return a Graph 

*/


    public Graph parseToStatement(String fn, GraphInterface gr)  {

	this.g=gr;
  
        if (fn.startsWith("file:")) {
            fn = fn.substring(5); 
        }


        //System.out.println("file: "+ fn);
	try{
        RandomAccessFile in = new RandomAccessFile(fn, "r");
        String text = "";
        String line = "";
        while (line != null){   
	    if(!line.startsWith("#") && (!line.equals(""))){
		    convertToStatement(line);
	    }
            line = in.readLine();
        }
	}catch(Exception e){
	System.err.println("NTParser error[3] "+e);
	}
    return (Graph)g;
     }


/** 

read in an Ntriples text file by filename and return the Graph passed to it

*/

    public Graph parseStringToStatement(String string, GraphInterface gr)  {

	g=gr;
	return parseStringToStatement(string);

    }


/** 

read in an Ntriples Stringand return a Graph 

*/

    public Graph parseStringToStatement(String string)  {
	try{

        StringTokenizer st = new StringTokenizer(string, "\n");
        boolean doneFirst = false;
        while (st.hasMoreTokens()) {
            String sss = st.nextToken();
	    convertToStatement(sss);
            doneFirst = true;  
        }
        
	}catch(Exception e){
	System.err.println("NTParser error[1] "+e);
	}

	return (Graph)g;
     }



/**

Tries to parse a String as a NTriples Triple; adds it to the
GraphInterface g.

*/

private void convertToStatement(String in){

		try{
		Statement q=processTripleToStatement(in);
			if(q==null){

			}else{
//			System.out.println("***adding statment "+q);
			g.tell(q);
			}
		}catch(Exception e){
		e.printStackTrace();
		}

}



/**

Nasty Ntriples parser, returning a Statement given a String (assumed to
be a line of NTriples)

*/


private Statement processTripleToStatement(String part){

Statement trip = null;

if(part!=null && (!part.trim().equals(""))){
part=part.trim();

trip=new Statement();


	int k=0;

        while(k<3){

        int i=part.indexOf(" ");

        if(k==2){
        i=part.lastIndexOf(".");
        }

	

	String node=part.substring(0,i).trim();

	part=part.substring(i+1).trim();

		if(node.startsWith("<")){
		Node r = Node.getResource(node.substring(1,node.lastIndexOf(">")));
			if(k==0){trip.setSubject(r);}
			if(k==1){trip.setPredicate(r);}
			if(k==2){trip.setObject(r);}
		}else 
		if(node.startsWith("_:")){
		String var=node.substring(node.indexOf("_:"));

			if (var.startsWith("?")){
			var=var.substring(1);

			((ComplexQuery)g).getVariables().addElement(var);
			}

			if(k==0){
			trip.setSubject(Node.getBlank(node));
			}
			if(k==1){
			trip.setPredicate(Node.getBlank(node));
			}
			if(k==2){
			trip.setObject(Node.getBlank(node));
			}
		}else 
		if(node.startsWith("\"")){
		int lastinvertedcomma=node.lastIndexOf("\"");
		Node r = Node.getLiteral(node.substring(1,node.lastIndexOf("\"")));
			if(k==0){System.out.println("error[1]");}
			if(k==1){System.out.println("error[2]");}
			if(k==2){trip.setObject(r);}
		}
		else if(node.startsWith("\'")){
		int lastinvertedcomma=node.lastIndexOf("\'");
		Node r = Node.getLiteral(node.substring(1,node.lastIndexOf("\"")));
			if(k==0){System.out.println("error[1]");}
			if(k==1){System.out.println("error[2]");}
			if(k==2){trip.setObject(r);}
		}

	k++;

	}

}
return trip;
}



}
