package org.ilrt.inkling.api;

import java.util.*;

/*
Copyright University of Bristol, 2003

This software is released under the W3C license:
http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231
*/


/**
<pre>
Inspired by http://www.w3.org/2001/12/rubyrdf/basicrdf.rb

This is a simple Node class. use Node.getResource(), Node.getLiteral()
and Node.getBlank() to create Nodes; also can have getSubjects(),
getObjects(), getPredicates(), with or without an identifier for the
Node as a  node-centric API.

There isn't any Caching or reuse of Nodes.

</pre>
*/

public class Node {

//the Graph to which the Node belongs
    Graph graph = new Graph(new Vector());

//the value of the Node
    String content = new String();

//an internal identifier
    String id = null;

//defaults to resource
    public boolean isresource = true;

//defaults to not bnode
    public boolean isbnode = false;

/**

Returns the Stringified value of the Node

*/
    public String getContent() {
        return content;
    }

/**

Returns the Graph which contains the Node

*/

    public Graph getGraph() {
        return graph;
    }

/**

Sets the content value of the Node

*/

    public void setContent(String content) {
        this.content = content;
    }


/**

For setting the Graph which contains the Node

*/
    public void setGraph(Graph graph) {
        this.graph = graph;
        //System.out.println("SETTING GRAPH"+graph.getID());
    }

/**

Returns the identifier of the Node

*/

    public String getID() {
        return id;
    }

/**

Sets the internal ID of the Node

*/

    public void setID(String id) {
        this.id = id;
    }


/**

Construtor for a basic Node with nothing in it

*/

    public Node() {
        String id = Util.getRandomString();
        this.id = id;
    }


/**

Construtor for a basic Node with specific content

*/
    public Node(String content) {
        this.content = content;
        this.id = Util.getRandomString();
    }

/**

Method for printing content

*/

    public String inspect() {
        return content;
    }

/**

Method for printing content

*/

    public String toString() {
        return inspect();
    }


/**

Use this method to create a Resource Node

*/
    public static Node getResource(String content) {
        return getResource(content, (Graph) Graph.getDefault());
    }


/**

Use this method to create a Resource Node, also setting the Graph that
contains it

*/


    public static Node getResource(String content, Graph graph) {

        //not sure how to handle that default graph thing
        if (graph == null) {
            graph = (Graph) Graph.getDefault();
        }

        //ignoring the local caching for now...?


        Node node = new Node(content);
        //if(content==null || content.equals("")){
        //System.out.println("Warning: node with no content created but not blank");
        //}
        node.setGraph(graph);
        return node;

    }

/**

Use this method to create a blank Node

*/

    public static Node getBlank(String id) {
        return getBlank(id, (Graph) Graph.getDefault());
    }

/**

Use this method to create a Blank Node, also setting the Graph that
contains it

Note that generation of ID where the source url is known is a hash of
the source with the id of the node from the parser, retaining
identity of Nodes

*/



    public static Node getBlank(String id, Graph graph) {

	String content;

        if (graph == null) {
            graph = (Graph) Graph.getDefault();
        }

        if (id == null || id.equals("")) {
            content = "_:"+Util.getRandomString();
        } else {

	      content = "_:"+Util.sha1(graph.getID()+id);
        }

        Node node = new Node(content);
        node.setGraph(graph);
        node.isbnode = true;
        return node;

    }


/**

Use this method to create a Literal Node

*/

    public static Node getLiteral(String content) {
        return getLiteral(content, (Graph) Graph.getDefault());
    }

/**

Use this method to create a Literal Node with a refernce to the Graph

*/



    public static Node getLiteral(String content, Graph graph) {

        if (graph == null) {
            graph = (Graph) Graph.getDefault();
        }

        if (content == null) {
            content = "";
        }

        Node node = new Node(content);
        node.setGraph(graph);
        node.isresource = false;
        return node;
    }

/**

Useful sha1 of content of Node

*/

    public String sha1() {
        String sha = "";
        try {
            sha = (new Integer(Util.hashcodeIntFromString(this.content))).
                    toString();
        } catch (java.security.NoSuchAlgorithmException e) {
            System.err.println("Node: could not create sha1 "+e);
        }
        return sha;
    }


/**

Node-centric Graph traversal

*/

    public Vector getObjects(String predid) {

        Query q = new Query( Node.getResource(this.getContent(),
                this.getGraph()),
                Node.getResource(predid, this.getGraph()), null);

        Graph g = this.getGraph().ask(q);

        //System.out.println("G"+this.getGraph().inspect());
        //System.out.println("Q"+q.inspect());

        return g.getObjects();

    }


/**

Node-centric Graph traversal

*/

    public Vector getObjects() {

        Query q = new Query( Node.getResource(this.getContent(),
                this.getGraph()), null, null);

        Graph g = this.getGraph().ask(q);

        //System.out.println("G"+this.getGraph().inspect());
        //System.out.println("Q"+q.inspect());

        return g.getObjects();

    }


/**

Node-centric Graph traversal

*/


    public Vector getSubjects(String predid) {

        Query q = new Query(null,
                Node.getResource(predid, this.getGraph()),
                Node.getResource(this.getContent(),
                this.getGraph()));

        Graph g = this.getGraph().ask(q);

        //System.out.println("G"+this.getGraph().inspect());
        //System.out.println("Q"+q.inspect());

        return g.getSubjects();

    }

/**

Node-centric Graph traversal

*/

    public Vector getSubjects() {

        Query q = new Query(null, null,
                Node.getResource(this.getContent(),
                this.getGraph()));

        Graph g = this.getGraph().ask(q);

        //System.out.println("G"+this.getGraph().inspect());
        //System.out.println("Q"+q.inspect());

        return g.getSubjects();

    }

/**

Node-centric Graph traversal

*/
    public Vector getPredicates(String objid) {

        Query q = new Query( Node.getResource(this.getContent(),
                this.getGraph()), null,
                Node.getResource(objid, this.getGraph()));

        Graph g = this.getGraph().ask(q);

        //System.out.println("G"+this.getGraph().inspect());
        //System.out.println("Q"+q.inspect());

        return g.getPredicates();

    }

/**

Node-centric Graph traversal

*/
    public Vector getPredicates() {

        Query q = new Query( Node.getResource(this.getContent(),
                this.getGraph()), null, null);

        Graph g = this.getGraph().ask(q);

        //System.out.println("G"+this.getGraph().inspect());
        //System.out.println("Q"+q.inspect());

        return g.getPredicates();

    }

    /**

Overriding equals for ease of comparison

     */

    public boolean equals(Object o) {

        Node n = (Node) o;
        String s2 = this.getContent();
        String s1 = n.getContent();

        if (s1.equals(s2)) {
            return true;
        } else {
            return false;
        }

    }

}
