package org.ilrt.inkling.api;

/*
Copyright University of Bristol, 2003

This software is released under the W3C license:
http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231
*/

/**
<pre>
A simple three-place query object. Basically a statement but with places
for variable names.
</pre>
*/

public class Query extends Statement {

    Node subject = null;
    Node predicate = null;
    Node object = null;

    String subjectVariable = null;
    String predicateVariable = null;
    String objectVariable = null;

    /** squishparser uses this */

    public Query() {
    }

/**
printing method

*/
    public String toString() {
        return inspect();
    }

/**
printing method

*/

    public String inspect() {
        return "QUERY:\n"+toNodeString(this.getSubject()) + "\n"+
                toNodeString(this.getPredicate()) + "\n"+
                toNodeString(this.getObject()) + "\n"+
                this.getSubjectVariable() + "\n"+
                this.getPredicateVariable() + "\n"+
                this.getObjectVariable() + "\n";

    }




/**

used by inspect()

*/

    public String toNodeString(Node n) {
        if (n == null) {
            return "novalue";
        } else {
            return n.toString();
        }

    }


/**

Constructor, calling a Statement

*/

    public Query(Node subject, Node predicate, Node object) {
        super(subject, predicate, object);
    }

/**

Constructor including variable names

*/

    public Query(Node subject, Node predicate, Node object,
            String svar, String pvar, String ovar) {
        super(subject, predicate, object);
        setSubjectVariable(svar);
        setPredicateVariable(pvar);
        setObjectVariable(ovar);
    }


/**

Accessor for subject variable name

*/
    public String getSubjectVariable() {
        return subjectVariable;
    }
/**

Accessor for predicate variable name

*/
    public String getPredicateVariable() {
        return predicateVariable;
    }

/**

Accessor for object variable name

*/

    public String getObjectVariable() {
        return objectVariable;
    }

/**

Setting subject variable name

*/

    public void setSubjectVariable(String var) {
        subjectVariable = var;
    }

/**

Setting predicate variable name

*/
    public void setPredicateVariable(String var) {
        predicateVariable = var;
    }

/**

Setting object variable name

*/

    public void setObjectVariable(String var) {
        objectVariable = var;
    }



/**

convert to a squish string: one way of querying sql database

*/
    public String toSquishString() {
        StringBuffer squish = new StringBuffer("select ");

        //poss bug in parser - optional ending ,. @@fixme

        if (subjectVariable != null && (!subjectVariable.equals(""))) {
            squish.append(" "+subjectVariable + ",");
        }
        if (predicateVariable != null && (!predicateVariable.equals(""))) {
            squish.append(" "+predicateVariable + ",");
        }
        if (objectVariable != null && (!objectVariable.equals(""))) {
            squish.append(" "+objectVariable);
        }

        squish.append(" where ("+getPredSquishContent() + " "+
                getSubSquishContent() + " "+getObjSquishContent() + ")");
        return squish.toString();

    }


/**

Used by toSquishString()

*/
    public String getSubSquishContent() {

        if (subject != null && subject.getContent() == null) {
            return subject.getContent();
        } else if (subjectVariable != null &&
                (!subjectVariable.equals(""))) {
            return subjectVariable;
        } else {
            return "s";
        }

    }

/**

Used by toSquishString()

*/
    public String getPredSquishContent() {

        if (predicate != null && predicate.getContent() == null) {
            return predicate.getContent();
        } else if (predicateVariable != null &&
                (!predicateVariable.equals(""))) {
            return predicateVariable;
        } else {
            return "p";
        }

    }

/**

Used by toSquishString()

*/

    public String getObjSquishContent() {

        if (object != null && object.getContent() == null) {
            return object.getContent();
        } else if (objectVariable != null && (!objectVariable.equals(""))) {
            return objectVariable;
        } else {
            return "o";
        }

    }



}
