package org.ilrt.inkling.api;

/*
Copyright University of Bristol, 2003

This software is released under the W3C license:
http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231
*/

/**
<pre>
Inspired by http://www.w3.org/2001/12/rubyrdf/basicrdf.rb

A simple Statement (Triple) object, with an ID. Except if created empty,
subject, predicate and object should not be null: use Query for simple
query objects with nulls.
</pre>
*/

public class Statement {

    Node subject;
    Node predicate;
    Node object;

/**

Empty constructor

*/

    public Statement() {
    }

/**

Constructor with Nodes. None should be null.

*/

    public Statement(Node subject, Node predicate, Node object) {
        this.subject = subject;
        this.predicate = predicate;
        this.object = object;
    }

/**

Identifier

*/

    public String getID() throws java.security.NoSuchAlgorithmException {
        return getSha1(subject, predicate, object);
    }

    private String getSha1(Node s, Node p,
            Node o) throws java.security.NoSuchAlgorithmException {
        String sid = "";
        String pid = "";
        String oid = "";
        String gid = "";

        if (s != null) {
            sid = s.getID();
            gid = s.getGraph().getID();
        }
        if (p != null) {
            pid = p.getID();
            gid = p.getGraph().getID();
        }
        if (o != null) {
            oid = o.getID();
            gid = o.getGraph().getID();
        }


        //return (new Integer(Util.hashcodeIntFromString(sid+""+pid+""+oid+""+gid))).toString();
        return ( new Integer(Util.hashcodeIntFromString(sid + pid + oid))
                ).toString();

    }


/**

Accessor

*/

    public String toString() {
        return inspect();
    }

/**

For printing/debugging

*/

    public String inspect() {
        return "Statement:\ns:"+subject.getID() + " p:"+
                predicate.getID() + " o:"+ object.getID() + " \ns:"+
                subject + " p:"+predicate + " o:"+
                object+" type "+object.isresource;
    }


/**

Set the subject Node.

*/

    public void setSubject(Node subject) {
        this.subject = subject;
    }

/**

Set the predicate Node.

*/

    public void setPredicate(Node predicate) {
        this.predicate = predicate;
    }

/**

Set the object Node.

*/

    public void setObject(Node object) {
        this.object = object;
    }

/**

Accessor for subject

*/

    public Node getSubject() {
        return subject;
    }

/**

Accessor for predicate

*/

    public Node getPredicate() {
        return predicate;
    }


/**

Accessor for object

*/

    public Node getObject() {
        return object;
    }


}
