package org.ilrt.inkling.app;

import org.ilrt.inkling.api.*;
import org.ilrt.inkling.query.*;
import java.util.*;
import java.io.*;

/*
Copyright University of Bristol, 2003

This software is released under the W3C license:
http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231
*/


/**
<pre>
This file tries to convert between different syntaxes for RDF Query, 
starting from an N-Triples query in queries/nt/, and exporting files to 
queries/ and to stdout. See http://www.w3.org/2001/11/13-RDF-Query-Rules/
</pre>
*/


public class DifferentSyntaxes {

static ComplexQuery cq= new ComplexQuery();
static String manifestFile="file:rdf/19-annotated-RDF-WSDL.rdf";

    public static void main(String args[]) {

		String filen="ericp";

		if(args!=null)
		{
			if(args.length==1)
			{

				String f=args[0];
				if(f.endsWith(".nt")){
					filen=f.substring(0,f.lastIndexOf(".nt"));
				}else{
					filen=f;
				}
			}

		}

		cq = new ComplexQuery(); 

                String url = "file:queries/nt/"+filen+".nt";
                String name = "syntaxes test";
                String desc = "Converting to different syntaxes - see Eric "+
"Prud'hommeaux's work at http://www.w3.org/2001/11/13-RDF-Query-Rules/";

		NTReader ntr2=new NTReader();
		Graph graph=ntr2.parseToStatement(manifestFile);
		cq.getGraphs().addElement(graph);

                System.out.println("\n\n"+name);
                System.out.println(desc);
                System.out.println("URL "+url);

			parseNT(url.trim());

			String algae=toAlgae();
	                System.out.println("\nALGAE version:\n"+
	                        algae);
			save(filen+".algae", algae);
	
			String N3=toN3();
	                System.out.println("\nN3 version:\n"+N3);
			save(filen+".n3", N3);
	
			String inkling=toInkling();
	                System.out.println("\nINKLING version:\n"+
	                        inkling);
			save(filen+".squish", inkling);
	
			String ruleml=toRuleML();
	                System.out.println("\nRuleML version:\n"+
	                        ruleml);
			save(filen+".ruleml", ruleml);
	
			String dql=toDQL();
 	                System.out.println("\nDQL version:\n"+
 	                       toDQL());
			save(filen+".dql", dql);
	                //not yet...			System.out.println("\nVersa version:\n"+toVersa());

		//}else{
		//	System.out.println("no query at "+filen+"nt");
		//}

    }//end method



    public static String toNT(String url, Vector graphs) {

        int i = url.lastIndexOf("/");
        String newUrl =
                url.substring(0, i + 1) + "nt"+url.substring(i) + ".nt";
        return newUrl;
    }


    public static void save(String filename, String content) {
        try {
            String s = filename;
            System.out.println("\n\nsaving to " + s);
            FileWriter fwr = new FileWriter(s, false);
            fwr.write(content);
            fwr.close();
        } catch (Exception e) {
            System.out.println("Problem saving " + e);
        }

    }




    public static String toAlgae(){

        String comment =
                "\n\nALGAE\nCan you have more than one graph? what does it look "+
                " like?\nAs it stands cannot pick variables, have shortened namespaces, as "+
                " well as not handling optional variables, constraints etc etc.\n\n";

        System.out.println(comment);

        String s = "(ask '";

        boolean source = false;
        boolean ns = false;

        //is graphs optional? can you have more than one? - assume one for now
        if (cq.getGraphs() != null) {

            Enumeration e = cq.getGraphs().elements();
            if (e.hasMoreElements()) {
                s = s + "("+e.nextElement().toString();
                source = true;
            }

        }

        Enumeration e = cq.getTriples().elements();
        while (e.hasMoreElements()) {

            Query p = (Query) e.nextElement();

            if (p.getPredicate() != null) {
                s = s + "\n\t("+p.getPredicate().getContent();
            } else {//can you have prd variables? probably...
                s = s + "\n\t(?"+p.getPredicateVariable();
            }

            if (p.getSubject() != null) {
                s = s + " "+p.getSubject().getContent();
            } else {
                s = s + " ?"+p.getSubjectVariable();
            }

            if (p.getObject() != null) {
                s = s + " "+p.getObject().getContent();
            } else {
                s = s + " ?"+p.getObjectVariable() + ")";
            }

        }

        if (source) {
            s = s + ")";
        }


        s = s + "\ncollect '(";


        Enumeration a = cq.getVariablesWeWant().elements();
        int si = cq.getVariablesWeWant().size();
        int i = 1;
        while (a.hasMoreElements()) {
            s = s + "?"+(String) a.nextElement();
            if (i < si) {
                s = s + " ";
            } else {
                s = s + "))";
            }
            i++;
        }

        return s;

    }



    public static String toN3(){

        String comment =
                "\n\nN3:\nThis conversion does not do shortened namespaces or "+
                "picking out the variables you want. I don't know if you can have source "+
                "graph(s) in N3.\n\n";

        System.out.println(comment);

        //prefixes not used

        String s = "@prefix agg: .\n\nthis log:forAll ";

        boolean source = false;
        boolean ns = false;


        Enumeration b = cq.getVariables().elements();
        int sii = cq.getVariables().size();
        int ii = 1;
        while (b.hasMoreElements()) {
            s = s + ":"+(String) b.nextElement();
            if (ii < sii) {
                s = s + ", ";
            } else {
                s = s + " . ";
            }
            ii++;
        }

        s = s + "\n{ ";



        Enumeration e = cq.getTriples().elements();
        while (e.hasMoreElements()) {
            Query p = (Query) e.nextElement();

            if (p.getSubject() != null) {
                s = s + "\n "+p.getSubject().getContent();
            } else {
                s = s + "\n :"+p.getSubjectVariable();
            }


            if (p.getPredicate() != null) {
                s = s + " "+p.getPredicate().getContent();
            } else {//can you have prd variables? probably...
                s = s + " :"+p.getPredicateVariable();
            }

            if (p.getObject() != null) {
                s = s + " "+p.getObject().getContent() + " .";
            } else {
                s = s + " :"+p.getObjectVariable() + " .";
            }

        }



        s = s + "\n} log:implies {";

        Enumeration a = cq.getVariablesWeWant().elements();
        int si = cq.getVariablesWeWant().size();
        int i = 1;
        while (a.hasMoreElements()) {
            s = s + "\n:"+(String) a.nextElement() + " a agg:QlAnswer .";
            if (i < si) {
                s = s + " . ";
            }
            i++;
        }

        return s + "\n} .";
    }



    public static String toInkling(){

        String comment =
                "\n\nInkling:\nThis conversion does not do shortened "+
                "namespaces or picking out the variables you want.\n\n";

        System.out.println(comment);

        String s = "SELECT ";

        Enumeration a = cq.getVariablesWeWant().elements();
        int si = cq.getVariablesWeWant().size();
        int i = 1;

        while (a.hasMoreElements()) {
            s = s + "?"+(String) a.nextElement();
            if (i < si) {
                s = s + ", ";
            }
            i++;
        }

        boolean source = false;
        boolean ns = false;

        //is graphs optional? can you have more than one? - assume one for now
        if (cq.getGraphs() != null) {

            int count = cq.getGraphs().size();
            int bla = 1;
            Enumeration e = cq.getGraphs().elements();
            while (e.hasMoreElements()) {
                s = s + "\nFROM "+e.nextElement().toString();
                if (count < bla) {
                    s = s + ", \n";
                }
            }

        }

        s = s + "\nWHERE ";

        Enumeration e = cq.getTriples().elements();

        while (e.hasMoreElements()) {

            Query p = (Query) e.nextElement();


            if (p.getPredicate() != null) {
                s = s + "\n   ("+p.getPredicate().getContent();
            } else {//can you have prd variables? probably...
                s = s + "\n   (?"+p.getPredicateVariable();
            }

            if (p.getSubject() != null) {
                s = s + " "+p.getSubject().getContent();
            } else {
                s = s + " ?"+p.getSubjectVariable();
            }

            if (p.getObject() != null) {
                s = s + " "+p.getObject().getContent();
            } else {
                s = s + " ?"+p.getObjectVariable() + ")";
            }

        }


        return s;
    }




    public static String toDQL(){

        String comment =
                "\n\nDQL:\nUses more up to date version of the syntax:"+
                " http://www.ksl.stanford.edu/projects/dql/syntax.shtml (from sandro). "+
                " http://www.daml.org/2002/08/dql/dql - can omit premise. "+
                "Not sure: if we can have more than one KB; if can do xxx, xox, xoo, oox.\n\n";

        System.out.println(comment);

        String dqvarns = "http://www.daml.org/2002/10/dql-variables#";


        String s =
                "\n<dql:query xmlns:dql=\"http://www.daml.org/2002/10/dql-syntax#\" "+
                "\n           xmlns:var=\"http://www.daml.org/2002/10/dql-variables#\" "+
                "\n           xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\" "+
                "\n           xmlns:rdfs=\"http://www.w3.org/2000/01/rdf-schema#\" "+
                "\n           xmlns:daml=\"http://www.daml.org/2001/03/daml+oil#\">";



        Enumeration e = cq.getTriples().elements();
        while (e.hasMoreElements()) {

            Query p = (Query) e.nextElement();

            String pred = null;
            String predVar = null;
            String sub = null;
            String subVar = null;
            String obj = null;
            String objVar = null;

            if (p.getPredicate() != null) {
                pred = p.getPredicate().getContent();
            } else {//can you have prd variables? maybe not...
                predVar = dqvarns + p.getPredicateVariable();
            }

            if (p.getSubject() != null) {
                sub = p.getSubject().getContent();
            } else {
                subVar = dqvarns + p.getSubjectVariable();
            }

            if (p.getObject() != null) {
                obj = p.getObject().getContent();
            } else {
                objVar = dqvarns + p.getObjectVariable();
            }

            /*
             can do oxo, xxo, oxx. anything else?

             */

            s = s + "\n  <dql:queryPattern>\n    <rdf:RDF>";

            if (sub == null && pred != null && obj == null) {
                s = s + "\n      <rdf:Description rdf:about=\""+subVar +
                        "\">";
                s = s + "\n         <"+pred + " rdf:resource=\""+
                        objVar + "\"/>";
            }

            if (sub != null && pred != null && obj == null) {
                s = s + "\n      <rdf:Description rdf:about=\""+sub + "\">";
                s = s + "\n         <"+pred + " rdf:resource=\""+
                        objVar + "\"/>";
            }

            if (sub == null && pred != null && obj != null) {
                s = s + "\n      <rdf:Description rdf:about=\""+subVar +
                        "\">";

                if (p.getObject().isresource) {
                    s = s + "\n     <"+pred + " rdf:resource=\""+obj + "\"/>";
                } else {
                    s = s + "\n     <"+pred + ">"+obj + "</"+pred + ">";
                }
            }


            s = s + "\n      </rdf:Description>\n    </rdf:RDF>\n  </dql:queryPattern>";


        }//end while


        Enumeration a = cq.getVariablesWeWant().elements();
        int si = cq.getVariablesWeWant().size();
        int i = 1;
        s = s + "\n<dql:mustBindVars>";

        while (a.hasMoreElements()) {
            s = s + "\n<"+dqvarns + (String) a.nextElement() + "/>";
            i++;
        }

        s = s + "\n</dql:mustBindVars>";

        if (cq.getGraphs() != null) {
            s = s + "\n<dql:answerKBPattern>";

            Enumeration ez = cq.getGraphs().elements();
            if (ez.hasMoreElements()) {
                s = s + "\n<dql:kbRef rdf:resource=\""+
                        ez.nextElement().toString() + "\"/>";
            }
            s = s + "\n</dql:answerKBPattern>";

        }

        s = s + "\n</dql:query>";
        return s;
    }


    ///ruleml....?

    public static String toRuleML(){


        String comment =
                "\n\nRuleML:\nThis conversion does not do shortened namespaces or "+
                "picking out the variables you want. Talking to Harold Boley, the "+
                "syntax, while acceptable, can be done better.\n\n";

        System.out.println(comment);

        String s = "";

        s = s + "<?xml version=\"1.0\" standalone=\"no\"?> "+ "<!DOCTYPE rulebase SYSTEM "+
                "\n\"http://www.dfki.de/ruleml/dtd/0.8/ruleml-urdatalog.dtd\"> "+
                "\n<rulebase>\n  <query>\n    <_body>\n      <and>";

        Enumeration e = cq.getTriples().elements();
        while (e.hasMoreElements()) {

            Query p = (Query) e.nextElement();


            if (p.getPredicate() != null) {
                s = s + "\n        <atom>\n         <_opr><rel href=\""+
                        p.getPredicate().getContent() + "\"><_opr>";
            } else {//can you have prd variables? probably...
                s = s + "\n         <var>"+p.getPredicateVariable() + "</var>";
            }

            if (p.getSubject() != null) {
                s = s + "\n         <ind href=\""+
                        p.getSubject().getContent() + "\"/>";
            } else {
                s = s + "\n         <var>"+p.getSubjectVariable() + "</var>";
            }

            if (p.getObject() != null) {
                s = s + "\n         <ind href=\""+
                        p.getObject().getContent() + "\"/>";
            } else {
                s = s + "\n         <var>"+p.getObjectVariable() + "</var>\n          </atom>";
            }

        }



        s = s + "\n      </and>\n    </_body>\n  </query>\n</rulebase>";

        return s;
    }


    ///versa

    /*

    distribute(type(rss:item),
               '. - rss:title -> *',
               '. - rss:link -> *')


    (type(rss:item) |-  rss:title -> contains('RDQL')) -> rss:link


    */



    public static String toVersa(){


        String comment =
                "\n\nVersa:\nI'm not sure about this. I've emailed Uche about "+
                "it.\n\n";

        System.out.println(comment);


        String s = "";
        s = s + "distribute(";

        Enumeration e = cq.getTriples().elements();
        while (e.hasMoreElements()) {

            Query p = (Query) e.nextElement();

            String pred = null;
            String predVar = null;
            String sub = null;
            String subVar = null;
            String obj = null;
            String objVar = null;

            if (p.getPredicate() != null) {
                pred = p.getPredicate().getContent();
            } else {//can you have prd variables? probably...
                predVar = p.getPredicateVariable();
            }

            if (p.getSubject() != null) {
                sub = p.getSubject().getContent();
            } else {
                subVar = p.getSubjectVariable();
            }

            if (p.getObject() != null) {
                obj = p.getObject().getContent();
            } else {
                objVar = p.getObjectVariable();
            }


            //using
            //http://uche.ogbuji.net:8080/uche.ogbuji.net/tech/rdf/versa/versa-by-example.txt

            //but how do we get what order they come in?


            //xxo
            if (sub != null && pred != null && obj == null) {
                s = s + "\n\" . - "+pred + " -> "+sub + "\",";
            }

            //oxx
            if (sub == null && pred != null && obj != null) {
                s = s + "\" . - "+pred + " <- "+obj + "\",";
            }
            /*??????
             //xox
             if(sub!=null && pred==null && obj!=null){
             s=s+"\n\" . - "+pred+" <- "+sub+"\",";
             }
             */
            //oox
            //xxo
            //oxo
            if (sub == null && pred != null && obj == null) {
                s = s + "\n\" . - "+pred + " -> * \",";
            }


        }

        s = s + ")";


        return s;
    }

    ///too hard....


public static void parseNT(String filen){


	if(filen.startsWith("file:")){
	filen=filen.substring(5);
	}


        NTReader ntr=new NTReader();
	cq=ntr.parseToQuery(filen);

}

}

