package org.ilrt.inkling.app;

import org.ilrt.inkling.api.*;
import java.util.*;

/*
Copyright University of Bristol, 2003

This software is released under the W3C license:
http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231
*/

/**
<pre>
Various useful schema-checking methods. See
http://swordfish.rdfweb.org/discovery/2003/08/validation/ for the online
demo.
</pre>
*/

public class Rosco{

//test values

static String schema="file:foaf.rdf";
static String testfile="file:libbyfoaf.rdf";
static String res="http://www.w3.org/2000/01/rdf-schema#Resource";

public static void main(String args[]){

//for each Person in schema, get all the properties which have 
//it as domain

String cl="http://xmlns.com/foaf/0.1/Person";

Rosco rama = new Rosco();
//rama.checkDomainConsistancy(cl, schema);
//rama.checkFileConsistancy(cl, schema, testfile);

Graph sc=new Graph(null,schema,Util.RDFXML);
sc.load();

//rama.getAllClasses(sc);

rama.fixSchemaDomains(sc);

}//end main


/**
Checks whether each property has a consistent domain with respect to a certain class
*/

public void checkDomainConsistancy(String cl, String schema){

System.out.println("this checks whether each property has a consistent domain with respect to a certain class.");
Graph sc=new Graph(null,schema,Util.RDFXML);
sc.load();
getProperties(sc,cl);

}


/**
This checks each node of a given type in a 
document, and lists the properties it could have, and does have, 
including superclass-related properties
*/

public void checkFileConsistancy(String cl,String schema,
String testfile){

//also extra properties?

System.out.println("This checks each node of a given type in a "+
"document, and lists the properties it could have, and does have, "+
"including superclass-related properties");

Graph test=new Graph(null,testfile,Util.RDFXML);
test.load();
Graph sc=new Graph(null,schema,Util.RDFXML);
sc.load();

Hashtable propsAndSource=new Hashtable();

Vector superclasses = new Vector();
superclasses.addElement(cl);
superclasses.addElement("http://www.w3.org/2000/01/rdf-schema#Resource");
getSuperClasses(sc,cl,superclasses);

Enumeration e = superclasses.elements();
	while(e.hasMoreElements()){
	String clss=(String)e.nextElement();
	Vector props=getPropertiesWithDomain(sc,clss);
		propsAndSource.put(clss,props);
	}

checkDocStatusByClass(test,cl,propsAndSource);


}

/**

For each class, and for a list of properties per class (in a Hash of
vectors, keyed by class name), checks if instances of the property are
present in Graph test.

*/

public void checkDocStatusByClass(Graph test,String cl,Hashtable properties){

System.out.println("CHECKING "+test.getID()+"\n");

//ok, check status of this document wrt cl

//e.g. for each Person, either
//- check wrt everything, or
//- check wrt a generated query

//get all Persons (is there a better, non-person-centric way?)

Vector classInstances =
test.askForSubjects(null,"http://www.w3.org/1999/02/22-rdf-syntax-ns#type",cl);


//for each of the classInstances:
//for each of the properties
//see if there's anything there in test

//System.out.println("people " +classInstances.toString() );
//System.out.println("properties " +properties.toString() );

Enumeration e = classInstances.elements();

	while(e.hasMoreElements()){
	Node n = (Node)e.nextElement();
		System.out.println("\nnode "+n.getContent());
		Enumeration e1=properties.keys();
		while(e1.hasMoreElements()){
			String key=(String)e1.nextElement();

			System.out.println("\tclass "+key);
			Vector myprops=(Vector)properties.get(key);

			Enumeration e2=myprops.elements();
			while(e2.hasMoreElements()){

			String prop=(String)e2.nextElement();
			Vector v =test.askForObjects(n,new Node(prop),null);
				if(v.size()>0){
				System.out.println("\t\ttrue "+prop+" "+v.size()+" "+v.elementAt(0));
				}else{
				System.out.println("\t\tfalse "+prop);
				}
			}

		}

	System.out.println("\n\n");

	}


}//end method


/**

gets all the properties with the domain of the class cl or any of its
superclasses

*/

public Vector getProperties(Graph schema, String cl){

Vector superclasses = new Vector();

superclasses.addElement(cl);
superclasses.addElement("http://www.w3.org/2000/01/rdf-schema#Resource");
getSuperClasses(schema,cl,superclasses);

Enumeration e = superclasses.elements();
	while(e.hasMoreElements()){
	String st = (String)e.nextElement();
	System.out.println(st);
	}

//get all props with this class as domain


Vector properties = new Vector();

Enumeration e2 = superclasses.elements();
	while(e2.hasMoreElements()){
	String ccll=(String)e2.nextElement();

System.out.println("properties with domain "+ccll);

	Vector props=getPropertiesWithDomain(schema,ccll);
	System.out.println(props.toString()+"\n");

	Enumeration e3=props.elements();
		while(e3.hasMoreElements()){
		properties.addElement((String)e3.nextElement());
		}
	}

//final check for things that don't have a domain

Vector allprops = getAllProperties(schema);

Enumeration e4=allprops.elements();

Vector missingprops=new Vector();

System.out.println("no domain:");

	while(e4.hasMoreElements()){
	String myprop=(String)e4.nextElement();
		if(!properties.contains(myprop)){
		missingprops.addElement(myprop);
		System.out.println(myprop);
		}
	}


return properties;

}

/**

Reursively gets all the superclasses of the class cl from the schema
schema, and adds to the Vector tofill.

*/

public void getSuperClasses(Graph schema, String cl, Vector tofill){

Vector 
v=schema.ask(cl,"http://www.w3.org/2000/01/rdf-schema#subClassOf",null);

Enumeration e = v.elements();
	while(e.hasMoreElements()){
	Statement st = (Statement)e.nextElement();
	//System.out.println(st);

	String ob=st.getObject().getContent();
		if(!tofill.contains(ob)){
//		System.out.println("adding "+ob);
		tofill.addElement(ob);
		getSuperClasses(schema,ob,tofill);
		}
	}


}//end method

/**

Recursively gets all the properties with domain cl from the schema.

*/

public Vector getPropertiesWithDomain(Graph schema, String cl){

Vector 
v=schema.ask(null,"http://www.w3.org/2000/01/rdf-schema#domain",cl);
Vector tofill=new Vector();

Enumeration e = v.elements();
	while(e.hasMoreElements()){
	Statement st = (Statement)e.nextElement();

	String sub=st.getSubject().getContent();
		if(!tofill.contains(sub)){
//		System.out.println("adding "+sub);
		tofill.addElement(sub);
		}

	}

return tofill;

}//end method
/**

Gets all the properties in a given schema, returning a unique vector of
Strings.

*/

public Vector getAllProperties(Graph schema){

Vector 
v=schema.ask(null,"http://www.w3.org/1999/02/22-rdf-syntax-ns#type","http://www.w3.org/1999/02/22-rdf-syntax-ns#Property");
Vector tofill=new Vector();

Enumeration e = v.elements();
	while(e.hasMoreElements()){
	Statement st = (Statement)e.nextElement();

	String sub=st.getSubject().getContent();
		if(!tofill.contains(sub)){
//		System.out.println("adding "+sub);
		tofill.addElement(sub);
		}

	}

return tofill;

}//end method


/**

Tests if a given s,p,o triples is present ina Graph.

*/

public boolean hasProperty(Graph g,Node s, Node p, Node o){

Vector v =g.ask(s,p,o);

if(v.size()>0){
return true;
}
else{
return false;
}

}


/**

Returns a list of Strings - all Classes in the schema.

*/

public Vector getAllClasses(Graph schema){

Vector 
v=schema.ask(null,"http://www.w3.org/1999/02/22-rdf-syntax-ns#type","http://www.w3.org/2000/01/rdf-schema#Class");
Vector tofill=new Vector();

Enumeration e = v.elements();
	while(e.hasMoreElements()){
	Statement st = (Statement)e.nextElement();

	String sub=st.getSubject().getContent();
		if(!tofill.contains(sub)){
		System.out.println("adding "+sub);
		tofill.addElement(sub);
		}

	}

return tofill;

}//end method

/**

For all properties in schema without a domain, adds a domain of
rdfs:Resource.

*/
public void fixSchemaDomains(Graph schema){

Vector 
v=schema.askForSubjects(null,"http://www.w3.org/1999/02/22-rdf-syntax-ns#type","http://www.w3.org/1999/02/22-rdf-syntax-ns#Property");
Enumeration e= v.elements();
	while(e.hasMoreElements()){

	Node n = (Node)e.nextElement();
	Node domain=new
Node("http://www.w3.org/2000/01/rdf-schema#domain");
	Vector v1=schema.ask(n,domain,
	null);

		if(v1.size()==0){
		schema.tell(new Statement(n,domain,new Node(res)));
		}

	}

}//end methid

/**

Adds an rdf:type rdfs:Resource for each subject in a graph.

*/


public void fixNoTypes(Graph graph){

Vector v=graph.getSubjects();

Enumeration e= v.elements();
	while(e.hasMoreElements()){
	Node n=(Node)e.nextElement();
	graph.tell(new Statement(n,new
Node("http://www.w3.org/1999/02/22-rdf-syntax-ns#type"),new Node(res)));
	}

}//end method


}//end class
