package org.ilrt.inkling.query;

import java.util.*;
import java.io.*;
import java.sql.*;
import java.net.*;
import org.ilrt.inkling.api.*;
import org.ilrt.inkling.external.*;

/*
Copyright University of Bristol, 2003

This software is released under the W3C license:
http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231
*/

/**
<pre>
A simple, in-memory query engine implementation.
In general, it should not be used directly, but via Graph, like this:

 	Graph gr = new Graph(null, uri, Util.RDFXML);
 	gr.load();
 	java.sql.ResultSet r=gr.askSquish(query);

or for squish queries with 'from' clauses specifying the urls to query, use

        Graph gr = new Graph();
        java.sql.ResultSet r=gr.askSquish(query);


or if you want to access the query (ComplexQuery) itself, do it like this:

        ComplexQuery cq=(ComplexQuery)SquishParser.parse(query);
        cq.getGraphs().addElement(ug);
        QE qe=new QE();

        java.sql.ResultSet r=qe.executeQuery(cq);

</pre>
*/



public class QE  {

static boolean beStrict=true;

public static void main(String args[]){


String filen="file:tests/rdql-tests-2003-04-10/queries/nt/test-4-04.nt";
String sourcefile="file:tests/rdql-tests-2003-04-10/inputs/model2.nt";

QE qe = new QE();
NTReader ntr1= new NTReader();
ComplexQuery complexQuery=ntr1.parseToQuery(filen);

Graph gr = new Graph(null, sourcefile, Util.NTRIPLES);

complexQuery.getGraphs().addElement(gr);

java.sql.ResultSet rs=qe.executeQuery(complexQuery);

System.out.println(((org.ilrt.inkling.query.ResultSet)rs).toString());

//eventually save
}

    String sourceHandler = "";
    Object source = null;//@@

    static int RESOURCE = 0;
    static int LITERAL = 1;

    //variables list
    public Vector variables = new Vector();

    //constraints list - should be Vectors of triples keyed off variable names
    public Hashtable constraints = new Hashtable();

    public QE(Object source, String handler) {
        this.source = source;
        sourceHandler = handler;
    }

    public QE() {
        this.source = null;
    }



    /**

     Makes a query of an RDF database and returns the results as a graph 
(GraphInterface)
     subject-, predicate- and object- variables are the variable names.
     subject, predicate and object can be null

     */


    public GraphInterface query(Query t,
            Object source) {

        SimpleQueryInterface database = null;
        GraphInterface results = new Graph(new Vector());
	Vector vv = new Vector();
            try {
                database = (SimpleQueryInterface) source;
                vv=database.askQuery(t.getSubject(), t.getPredicate(), t.getObject());

            } catch (Exception e) {
                System.out.println("Query error[1] " + e);
            System.out.println("\n\nprob: " + e + " sub " +
                    t.getSubject() + " pred " + t.getPredicate() +
                    " obj " + t.getObject() + "\n\n");
                e.printStackTrace();
            }

	results.tellAll(vv);
        return results;
    }



    /**

        builds a tree, flattening out the graph.

        what happens is this.
        for every query, you need to pass it a GraphInterface, and a triple constaining the
        variable names in its Variables.
        this method creates a tree out of TreeNode s, which have text as their varlue, and
        var as their variable name. so they're like variable-result pairs

        this method finds alol the subjects in a grph, and then for all the subjects finds
        the predicate-object pair associated with it.

        if this is the first set of results it just adds the subject nodes to a special
        'top' node, and then adds the predicate nodes to the subject nodes and the object
        nodes to the prediacte nodes. so you get

        top		*
         	      /   \
        S	     *     *
      		   /  \   / \
        P	  *    * *   *
         	 /     \ /    \
        O

        e.g. 	top

        	http://image/bla

        	depicts

        	person1		person2


        each layer can have as many branches as necessary, e.g often there will be only one
        predicate layer (from each of the subject nodes).

        Then for subsequant layers, we call hasTextRecursive on a list of the bottom nodes
        in the tree. for each of subsquent s,p,o, in the resultsets, we work out if the
        text assciated with the node is in the current branch. if it is, we add to the
        branch.

        e.g. what do we do with the result triple

        http://image/bla description "great image!"

        http://image/bla is in both branches, so we add 'description' and 'great image!'
        nodes to both branches

        e.g. 	top

        	http://image/bla

        	depicts

        	person1		person2

        	description	description

        	"great image!" 	"great image!"


        this means that we always get a tree not a graph; we can build the resultset by
        starting from the bottom nodes and getting the parent recusively.


        */

    Vector bottoms = new Vector();
    TreeNode top;
    boolean firstround = true;


    public void createIndex(GraphInterface results1,
            Query t1) {


        Vector newBottoms = new Vector();

        //it's the first round, so we add a special top node


        if (bottoms.size() == 0 && firstround) {
//        if (bottoms.size() == 0) {
            top = new TreeNode("top");
            bottoms.addElement(top);
            top.setVar("top");
//            top.setType(RESOURCE);
        }


        Vector dd = results1.getStatements();

        Enumeration eee = dd.elements();

        while (eee.hasMoreElements()) {

            org.ilrt.inkling.api.Statement t =
                    (org.ilrt.inkling.api.Statement) eee.nextElement();


            //loop through the bottom nodes, since we need to add a copy
            //of everything to each relevant branch

            Enumeration bots = bottoms.elements();


            test:
            while (bots.hasMoreElements()) {

                TreeNode bottom = (TreeNode) bots.nextElement();

		TreeNode pred = new TreeNode(t.getPredicate());
		TreeNode sub = new TreeNode(t.getSubject());
		TreeNode obj = new TreeNode(t.getObject());

                pred.setVar(t1.getPredicateVariable());
                sub.setVar(t1.getSubjectVariable());
                obj.setVar(t1.getObjectVariable());


                /*
                 start of constraints part.
                 if the constraint fails we fall out of this loop and go to the next bottom
                 */

                if (constraints.containsKey(t1.getPredicateVariable())) {


                    Vector vv = (Vector) constraints.get(
                            t1.getPredicateVariable());

                    Enumeration re = vv.elements();

                    while (re.hasMoreElements()) {

                        Query constraint =
                                (Query)
                                re.nextElement();

                        //does the constraint checking
                        if (constraintCheck(pred, constraint)) {
                            continue;
                        } else {
                            break test;
                        }
                    }
                }



                if (constraints.containsKey(t1.getSubjectVariable())) {
                    //check constraint ok

                    Vector vv = (Vector) constraints.get(
                            t1.getSubjectVariable());

                    Enumeration re = vv.elements();

                    while (re.hasMoreElements()) {

                        Query constraint =
                                (Query)
                                re.nextElement();

                        if (constraintCheck(sub, constraint)) {
                            continue;
                        } else {
                            break test;
                        }
                    }
                }



                if (constraints.containsKey(t1.getObjectVariable())) {
                    //check constraint ok

                    Vector vv = (Vector) constraints.get(
                            t1.getObjectVariable());

                    Enumeration re = vv.elements();

                    while (re.hasMoreElements()) {

                        Query constraint =
                                (Query)
                                re.nextElement();

                        if (constraintCheck(obj, constraint)) {
                            continue;
                        } else {
                            break test;
                        }
                    }
                }




                ///end constraint check

                int subB = bottom.hasTextVarRecursive(sub.getContent(),
                        sub.getVar(),sub.isresource);
                int predB = bottom.hasTextVarRecursive(pred.getContent(),
                        pred.getVar(),pred.isresource);
                int objB = bottom.hasTextVarRecursive(obj.getContent(),
                        obj.getVar(),obj.isresource);

//		System.out.println("INTS are "+subB+" "+predB+" "+objB);


                if (subB == 0 & predB == 0 & objB == 0) {
                    //dont add anything
		    ///hm.....
                } else if (subB == 0 & predB == 0 & objB == 1) {
                    bottom.add(obj);

                    newBottoms.addElement(obj);

                } else if (subB == 0 & predB == 1 & objB == 0) {
                    bottom.add(pred);

                    newBottoms.addElement(pred);
                } else if (subB == 1 & predB == 0 & objB == 0) {
                    bottom.add(sub);
                    newBottoms.addElement(sub);

                } else if (subB == 0 & predB == 1 & objB == 1) {

                    bottom.add(pred);
                    pred.add(obj);

                    newBottoms.addElement(obj);

                } else if (subB == 1 & predB == 1 & objB == 0) {

                    bottom.add(pred);
                    pred.add(sub);
                    newBottoms.addElement(sub);
                } else if (subB == 1 & predB == 0 & objB == 1) {
                    bottom.add(obj);
                    obj.add(sub);

                    newBottoms.addElement(sub);

                } else if (subB == 1 & predB == 1 & objB == 1) {
//hm
//                    if (firstround) {

                        bottom.add(sub);
                        sub.add(pred);
                        pred.add(obj);

                        newBottoms.addElement(obj);
//                    }


                }

            }

        }


        firstround = false;
        bottoms = newBottoms;
//System.out.println("GOT nodes bottom: "+bottoms);


    }//end method



    /*
    Given a vector of Queries, returns a vector of the variables they contain
    */

    public Vector listVariables(Vector triples) {
        Vector v = new Vector();
        Enumeration tr = triples.elements();
        while (tr.hasMoreElements()) {
            Query t =
                    (Query) tr.nextElement();
            if (t.getSubjectVariable() != null) {
                if (t.getSubjectVariable().toString().startsWith("?") &&
                        (!v.contains(t.getSubjectVariable()))) {
                    v.addElement(t.getSubjectVariable());
                }
            }
            if (t.getObjectVariable() != null) {
                if (t.getObjectVariable().toString().startsWith("?") &&
                        (!v.contains(t.getObjectVariable()))) {
                    v.addElement(t.getObjectVariable());
                }
            }
            if (t.getPredicateVariable() != null) {
                if (t.getPredicateVariable().toString().startsWith("?") &&
                        (!v.contains(t.getPredicateVariable()))) {
                    v.addElement(t.getPredicateVariable());
                }
            }
        }
        return v;
    }

/**

A list of all the variables used

*/

    public Vector listVariablesRaw(Vector triples) {
        Vector v = new Vector();
        Enumeration tr = triples.elements();
        while (tr.hasMoreElements()) {
            Query t =
                    (Query) tr.nextElement();
            if (t.getSubject() != null) {
                t.setSubject(t.getSubject());
                if (t.getSubject().toString().startsWith("?") &&
                        (!v.contains(t.getSubject()))) {
                    v.addElement(t.getSubject());
                }
            }
            if (t.getObject() != null) {
                t.setObject(t.getObject());
                if (t.getObject().toString().startsWith("?") &&
                        (!v.contains(t.getObject()))) {
                    v.addElement(t.getObject());
                }
            }
            if (t.getPredicate() != null) {
                t.setPredicate(t.getPredicate());
                if (t.getPredicate().toString().startsWith("?") &&
                        (!v.contains(t.getPredicate()))) {
                    v.addElement(t.getPredicate());
                }
            }
        }
        return v;
    }


   	 /**

	executeQuery is called with a ComplexQuery object
	this is basically a graph but with variables and optionally
	constraints and 'graphs' ('from' stuff)

	It returns a java.sql.Resultset. 
	@@sql part? not needed now?

	 */

    public java.sql.ResultSet executeQuery(ComplexQuery cq) {

        java.sql.ResultSet rows = new org.ilrt.inkling.query.ResultSet();


        bottoms = new Vector();

        variables = cq.getVariablesWeWant();

	if((variables.size()==0) && (!beStrict)){
		variables = cq.getVariables();
	}

	//System.out.println("VARS "+variables);

        top = null;
        firstround = true;

        fillConstraints(cq.getConstraints());
        Vector v = cq.getTriples();

        //*****// SQL


        if (source instanceof org.ilrt.inkling.query.QueryDirect) {


            QueryDirect qd = (QueryDirect) source;

            //SquishParser
            try {

                rows = qd.makeQuery(cq);

            } catch (Exception er) {
                //ErrorLog.write("QE: could not make query " + er);
            }
        } else {

            //*****//not SQL

            if (cq.getGraphs().size() != 0) {
                sourceHandler = "org.ilrt.inkling.api.Graph";

		if(source==null){
			source=new Graph();
		}

//not sure about this

		 for (Iterator iterator = cq.getGraphs().listIterator();
                        iterator.hasNext();) {
			Graph g=(Graph) iterator.next();

			if(g.size()==0){
			g.load();
			}
			((Graph)source).tellAll(g.getStatements());
 		 }

            }else{
                System.out.println("OOPS "+cq.getGraphs());

	    }

            //now loop through the query triples making the query and creating the index.

	//System.out.println("VVV "+v.size());

            Enumeration e = v.elements();
            while (e.hasMoreElements()) {

                Query triple =
                        (Query) e.nextElement();
		

                GraphInterface results = query(triple, (Object) source);

                createIndex(results, triple);

            }


            /**

             just loops through the list of bottoms creating a row for each one
             by calling toResultSet, which is resursive

             */

            rows = new org.ilrt.inkling.query.ResultSet();

            Enumeration rr = bottoms.elements();


            int i = 0;

            while (rr.hasMoreElements()) {

                Hashtable row = new Hashtable();

              //System.out.println("\nrow ["+i + "]");
                i++;

//TreeNode tmp=(TreeNode)rr.nextElement();
                toResultSet((TreeNode) rr.nextElement(), row);
              //System.out.println("\nrow ["+row + "]");

                ((org.ilrt.inkling.query.ResultSet) rows).addElement(
                        (Object) row);
            }

        }

        return rows;
    }



    /**

        little recursive method for climbing the resultset tree

        */

    public void toResultSet(TreeNode tn, Hashtable row) {

        //assume at the bottom
	//there seems to be an issue with it putting unwanted variables in
	//need to ID them....with a "?"
	//??need to fix for variableswewant


        if (variables.size() != 0 && variables.contains(tn.getVar()) ) {


//this makes a huge object: BUT - good to have the actual one?
//e.g. for get graph, get node...
	            row.put(tn.getVar(), (Node)tn);

	}

        TreeNode parent = tn.getParent();


        if (parent != null) {
            toResultSet(parent, row);
        }

    }


    /**

        constraints method - gets a constraint from a triple and returns a boolean if it
        matches. little bit rough around the edges but seems to work ok.

        */



    public boolean constraintCheck(TreeNode tn,
            Query t) {


        //should be something like ?x < 4
        String s = t.getSubject().toString();
        String p = t.getPredicate().toString();
        String o = t.getObject().toString();

        String text = tn.getContent();
        String val = tn.getVar();

        /*
        	- eq
        	- ne
        	- <,>,<=,>=,=,!=
        	- like, ~
        */

        //cast to string?

        try {

            if (p.equals("eq")) {

                //exact (case sensitive) match
                if (s.equals(val) && text.equals(o)) {
                    return true;
                } else {
                    return false;
                }

            } else if (p.equals("ne")) {


                //exact (case sensitive) match
                if (s.equals(val) && (!text.equals(o))) {

                    return true;
                } else {

                    return false;
                }


            } else if (p.equals("like") || p.equals("~")) {



                if (text.toLowerCase().indexOf(o.toLowerCase()) != -1) {
                    return true;
                } else {

                    return false;
                }
            }

            //cast to int/float?

            else if (p.equals(">")) {
                if ((new Integer(text)).intValue() >
                        (new Integer(o)).intValue()) {
                    return true;
                } else {
                    return false;
                }

            } else if (p.equals("<")) {
                if ((new Integer(text)).intValue() <
                        (new Integer(o)).intValue()) {
                    return true;
                } else {
                    return false;
                }
            } else if (p.equals(">=")) {
                if ((new Integer(text)).intValue() >=
                        (new Integer(o)).intValue()) {
                    return true;
                } else {
                    return false;
                }
            } else if (p.equals("<=")) {
                if ((new Integer(text)).intValue() <=
                        (new Integer(o)).intValue()) {
                    return true;
                } else {
                    return false;
                }

            } else if (p.equals("=") || p.equals("==")) {
                if ((new Integer(text)).intValue() ==
                        (new Integer(o)).intValue()) {
                    return true;
                } else {
                    return false;
                }
            } else if (p.equals("!=")) {



                if ((new Integer(text)).intValue() !=
                        (new Integer(o)).intValue()) {
                    return true;
                } else {
                    return false;
                }
            } else {
                return false;
            }
        }
        catch (java.lang.NumberFormatException nfe) {
            System.out.println("QE: not an integer - error");
            return false;
        }

    }



    /**

     Just puts constraints in a more useful format

     */

    public void fillConstraints(Vector constraints) {

        this.constraints = new Hashtable();
        Enumeration e = constraints.elements();

        while (e.hasMoreElements()) {


            Query t =
                    (Query) e.nextElement();

            String sub = t.getSubject().toString();

            Vector v = null;
            if (this.constraints.containsKey(sub)) {
                v = (Vector) this.constraints.get(sub);
            } else {
                v = new Vector();
                this.constraints.put(sub, v);
            }


            v.addElement(t);


        }

    }


}


