package org.ilrt.inkling.query;

import org.ilrt.inkling.api.*;
import java.sql.*;
import java.util.*;

/*
Copyright University of Bristol, 2003

This software is released under the W3C license:
http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231
*/

/**
<pre>
An implementation of java.sql.ResultSet. Many methods are not implemented.

There are several extra methods:

getColumnNames() - returns a Vector of columnnames

public Node getNode(int columnIndex)  - returns an SQLNode, creating 
bnode and literal values from their syntactic flags in the database.

public Node getNode(String columnName)  returns an SQLNode, creating
bnode and literal values from their syntactic flags in the database.


    public String toRDFString() - for RDF resultset. see  
http://www.w3.org/2003/03/rdfqr-tests/recording-query-results.html
http://www.w3.org/2003/03/rdfqr-tests/


The idea is that you can either return the ResultSet directly from the 
SQL database OR process it to get SQLNode Objects back.


Use it like this:

        SQLGraph sq = new SQLGraph();
        sq.setdb(db);
        sq.setDriver("com.mysql.jdbc.Driver");
        Graph gr = new Graph(null, uri, Util.RDFXML);
        gr.load();
        sq.add(gr);

        java.sql.ResultSet r1=sq.askSquish(query);
	org.ilrt.inkling.query.ResultSet r2=(org.ilrt.inkling.query.ResultSet)r1;

        SQLNode mbsh1=(Node)r2.getSQLNode("mbsh1");

using 

        String mbsh1=(Node)r1.getString("mbsh1");

is also fine, although you will need to post-process literals (which start 
with ")

@@provenance
</pre>
*/

public class ResultSet implements java.sql.ResultSet {

//the 'real' ResultSet, returned directly from the SQL database
java.sql.ResultSet resset=null;

/**

Used when accessing the SQL ResultSet directly

*/

public void setResultSet(java.sql.ResultSet r){
resset=r;
}


    int currentRow = 0;

    boolean sqlprov=false;
    Vector container= new Vector();

    String errorMessage = "oops!";

    Vector columns = new Vector();
    Vector colnames = new Vector();
    boolean showMsgs = false;

    /** 

overloaded method name to make indexing by column easier 

*/

    public void addElement(Object o) {
        Hashtable cleanRows = new Hashtable();

        if (o instanceof java.util.Hashtable) {
            Enumeration e = ((java.util.Hashtable) o).keys();
            while (e.hasMoreElements()) {
                String key = (String) e.nextElement();
                Object val = ((java.util.Hashtable) o).get(key);


                if (key.startsWith("?")) {
                    key = key.substring(1);
                }

                if (!columns.contains(val)) {
                    columns.addElement(val);
                }
                if (!colnames.contains(key)) {
                    colnames.addElement(key);
                }

                cleanRows.put(key, val);
            }
            container.addElement(cleanRows);

        }
    }

/**

An extra convenience method to access a Vector of column names

*/

    public Vector getColumnNames() {
        return colnames;
    }


/**

A method to set whether it tries to get provenance or not (SQL provenace
results in duplicates at the moment

*/

public void setSQLProv(boolean b){
sqlprov=b;
}


/**

java.sql.ResultSet method, either accessing the SQL ResultSet directly, 
or the in-memory version if that is not present.

*/

    public boolean next() throws SQLException {

if(resset!=null){
return resset.next();
}else{

        //        if (currentRow <= (this.size())) {
        if (currentRow < (container.size())) {
            currentRow++;
            return true;
        } else {
            currentRow = 0;
            return false;
        }
}
    }


    /**
 Results should always be stored as variables without ?, but can be 
accessed either way.
      */

    public String getString(String columnName) throws SQLException {

//System.out.println("YOU ARE HERE");

	if(resset!=null){
		String s =(String)resset.getString(columnName);
		if(s!=null && s.startsWith("\"")){
		s=s.substring(1);
		}

	return resset.getString(columnName);
	}else{

	        if (columnName.startsWith("?")) {
	            columnName = columnName.substring(1);
	        }
	        Hashtable row = null;
	        if (currentRow <= container.size()) {
	            row = (Hashtable) container.elementAt(currentRow - 1);
	        }
	        if (row != null) {
//???why not node?   - because want to access the actual node? 
///hm	TreeNode s = (TreeNode) row.get(columnName);
	            Node s = (Node) row.get(columnName);
//            String s = (String) row.get(columnName);
//            if (s != null && (!s.equals("null"))) {
        	    if (s != null){
        	        return s.getContent();
        	    } else {
                //                return errorMessage;
        	        return null;
	            }
	        } else {
            if (showMsgs) {
                return "\nend of results\n";
            } else {
                return null;
            }
        }

}

    }

/**

java.sql.ResultSet method

*/

    public void close() throws SQLException {
        container.removeAllElements();
    }


    public void setErrorMessage(String msg) {
        errorMessage = msg;
    }

    public String getErrorMessage() {
        return errorMessage;
    }

/**

Not implemented

*/

    public String getString(int columnIndex) throws SQLException {
        return null;
    }

/**

Not implemented

*/
    public boolean wasNull() throws SQLException {
        return false;
    }

/**

Not implemented

*/
    //nothing yet
    public boolean getBoolean(int columnIndex) throws SQLException {
        return false;
    }

/**

Not implemented

*/
    //nothing yet
    public byte getByte(int columnIndex) throws SQLException {
        return 0;
    }

/**

Not implemented

*/
    //nothing yet
    public short getShort(int columnIndex) throws SQLException {
        return 0;
    }

/**

Not implemented

*/
    //nothing yet
    public int getInt(int columnIndex) throws SQLException {
        return 0;
    }

/**

Not implemented

*/
    //nothing yet
    public java.net.URL getURL(int columnIndex) throws SQLException {
        return null;
    }


/**

Not implemented

*/
    //nothing yet
    public java.net.URL getURL(String url) throws SQLException {
        return null;
    }

/**

Not implemented

*/
    //nothing yet
    public long getLong(int columnIndex) throws SQLException {
        return 0;
    }

/**

Not implemented

*/
    //nothing yet
    public float getFloat(int columnIndex) throws SQLException {
        return 0;
    }

/**

Not implemented

*/
    //nothing yet
    public double getDouble(int columnIndex) throws SQLException {
        return 0;
    }

/**

Not implemented

*/
    //nothing yet
    public java.math.BigDecimal getBigDecimal(int columnIndex,
            int scale) throws SQLException {
        return null;
    }

/**

Not implemented

*/
    //nothing yet
    public byte[] getBytes(int columnIndex) throws SQLException {
        return null;
    }

/**

Not implemented

*/
    //nothing yet
    public java.sql.Date getDate(int columnIndex) throws SQLException {
        return null;
    }

/**

Not implemented

*/
    //nothing yet
    public Time getTime(int columnIndex) throws SQLException {
        return null;
    }

/**

Not implemented

*/
    //nothing yet
    public Timestamp getTimestamp(int columnIndex) throws SQLException {
        return null;
    }

/**

Not implemented

*/
    //nothing yet
    public java.io.InputStream getAsciiStream(int columnIndex)
            throws SQLException {
        return null;
    }

/**

Not implemented

*/

    //nothing yet
    public java.io.InputStream getUnicodeStream(int columnIndex)
            throws SQLException {
        return null;
    }

/**

Not implemented

*/
    //nothing yet
    public java.io.InputStream getBinaryStream(int columnIndex)
            throws SQLException {
        return null;
    }

/**

Not implemented

*/
    //nothing yet
    public boolean getBoolean(String columnName) throws SQLException {
        return false;
    }

/**

Not implemented

*/
    //nothing yet
    public byte getByte(String columnName) throws SQLException {
        return 0;
    }
/**

Not implemented

*/

    //nothing yet
    public short getShort(String columnName) throws SQLException {
        return 0;
    }
/**

Not implemented

*/

    //nothing yet
    public int getInt(String columnName) throws SQLException {
        return 0;
    }

/**

Not implemented

*/
    //nothing yet
    public long getLong(String columnName) throws SQLException {
        return 0;
    }

/**

Not implemented

*/
    //nothing yet
    public float getFloat(String columnName) throws SQLException {
        return 0;
    }

/**

Not implemented

*/
    //nothing yet
    public double getDouble(String columnName) throws SQLException {
        return 0;
    }

/**

Not implemented

*/
    //nothing yet
    public java.math.BigDecimal getBigDecimal(String columnName,
            int scale) throws SQLException {
        return null;
    }
/**

Not implemented

*/

    //nothing yet
    public byte[] getBytes(String columnName) throws SQLException {
        return null;
    }
/**

Not implemented

*/

    //nothing yet
    public java.sql.Date getDate(String columnName) throws SQLException {
        return null;
    }
/**

Not implemented

*/

    //nothing yet
    public Time getTime(String columnName) throws SQLException {
        return null;
    }
/**

Not implemented

*/

    //nothing yet
    public Timestamp getTimestamp(String columnName) throws SQLException {
        return null;
    }
/**

Not implemented

*/

    //nothing yet
    public java.io.InputStream getAsciiStream(String columnName)
            throws SQLException {
        return null;
    }
/**

Not implemented

*/

    //nothing yet
    public java.io.InputStream getUnicodeStream(String columnName)
            throws SQLException {
        return null;
    }

/**

Not implemented

*/
    //nothing yet
    public java.io.InputStream getBinaryStream(String columnName)
            throws SQLException {
        return null;
    }
/**

Not implemented

*/

    //nothing yet
    public SQLWarning getWarnings() throws SQLException {
        return null;
    }
/**

Not implemented

*/

    //nothing yet
    public void clearWarnings() throws SQLException {
    }

/**

Not implemented

*/
    //nothing yet
    public String getCursorName() throws SQLException {
        return null;
    }


/**
returns either the SQL ResultSet's ResultSetMetaData object, or an 
in-memory implementation for the in-memory version.

*/
    public java.sql.ResultSetMetaData getMetaData() throws SQLException {


        java.sql.ResultSetMetaData r =null;
if(resset!=null){
r=resset.getMetaData();
}else{
                new org.ilrt.inkling.query.ResultSetMetaData(this);
}
        return r;
    }


/**

Either form the SQL ResultSet or from the  
in-memory implementation for the in-memory version.

*/

    public Object getObject(int columnIndex) throws SQLException {

if(resset!=null){
return resset.getObject(columnIndex);
}else{
        String columnName = (String) colnames.elementAt(columnIndex);
        if (columnName != null) {
            return getObject(columnName);
        } else {
            return null;
        }
}
    }

/**

Returns an SQLNode if available, or creates one if not.

*/

    public Node getNode(int columnIndex) throws SQLException {

//@@fixme - prv, lit
	Object o=getObject(columnIndex);

	if(o instanceof SQLNode){
	return (SQLNode)o;
	}else{
	return new SQLNode((String)o);
	}
    }

/**

Returns an SQLNode if available, or creates one if not.

*/

    public Node getNode(String columnName) throws SQLException {

	Object o=getObject(columnName);
	String id=null;//sqlprov defaults to false
	if(sqlprov){id=getString(columnName+"_id");}

	if(o instanceof SQLNode){
	return (Node)o;
	}else{

//for inmem/sql
	
		if(o instanceof org.ilrt.inkling.query.TreeNode){
		return (Node)o;
		}else{
		SQLNode n = new SQLNode((String)o);

//hmm
//still need literals, bnodes etc
		n.setGraph(new Graph(null,id));
		return (SQLNode)n;
		}
	}

    }


/**

Either form the SQL ResultSet or from the  
in-memory implementation for the in-memory version.

*/


    public Object getObject(String columnName) throws SQLException {

//System.out.println("GOT col n "+columnName);

if(resset!=null){
return resset.getObject(columnName);
}else{

        if (columnName.startsWith("?")) {
            columnName = columnName.substring(1);
        }
        Hashtable row = null;
        if (currentRow <= container.size()) {
            row = (Hashtable) container.elementAt(currentRow - 1);
        }
        if (row != null) {
            Object s = row.get(columnName);
//System.out.println("row is "+row);

            if (s != null) {
                return s;
            } else {
                //                return errorMessage;
                return null;
            }
        } else {
//System.out.println("row is null ");
            if (showMsgs) {
                return "\nend of results\n";
            } else {
                return null;
            }
        }

}
    }


/**

Not implemented

*/
    //nothing yet
    public int findColumn(String columnName) throws SQLException {
        return 0;
    }

/**

Not implemented

*/
    public void cancelRowUpdates() {
    }

/**

Not implemented

*/
    public void deleteRow() {
    }

/**

Not implemented

*/
    public void updateRow() {
    }

/**

Not implemented

*/
    public void refreshRow() {
    }

/**

Not implemented

*/
    public java.io.Reader getCharacterStream(int i) {
        return null;
    }

/**

Not implemented

*/
    public java.io.Reader getCharacterStream(java.lang.String s) {
        return null;
    }

/**

Not implemented

*/
    public void updateDouble(int i, double d) {
    }

/**

Not implemented

*/
    public void updateDouble(java.lang.String s, double d) {
    }

/**

Not implemented

*/
    public java.sql.Time getTime(int i, java.util.Calendar c) {
        return null;
    }

/**

Not implemented

*/
    public java.sql.Time getTime(java.lang.String s, java.util.Calendar c) {
        return null;
    }

/**

Not implemented

*/
    public int getConcurrency() {
        return 0;
    }

/**

Not implemented

*/
    public void moveToInsertRow() {
    }

/**

Not implemented

*/
    public void updateDate(int i, java.sql.Date d) {
    }

/**

Not implemented

*/
    public void updateDate(java.lang.String s, java.sql.Date d) {
    }

/**

Not implemented

*/
    public int getFetchSize() {
        return container.size();
    }

/**

Not implemented

*/
    public void updateBytes(int i, byte[] b) {
    }

/**

Not implemented

*/
    public void updateBytes(java.lang.String s, byte[] b) {
    }

/**

Not implemented

*/

    public int getFetchDirection() {
        return 0;
    }

/**

Not implemented

*/
    public boolean relative(int i) {
        return false;
    }

/**

Not implemented

*/
    public void updateByte(int i, byte b) {
    }

/**

Not implemented

*/
    public void updateByte(java.lang.String s, byte b) {
    }

/**

Not implemented

*/
    public void updateCharacterStream(int i, java.io.Reader r, int j) {
    }

/**

Not implemented

*/
    public void updateCharacterStream(java.lang.String s,
            java.io.Reader r, int j) {
    }

/**

Not implemented

*/
    public java.sql.Date getDate(int i, java.util.Calendar c) {
        return null;
    }

/**

Not implemented

*/
    public java.sql.Date getDate(java.lang.String s, java.util.Calendar c) {
        return null;
    }

/**

Not implemented

*/
    public boolean rowInserted() {
        return false;
    }

/**

Not implemented

*/
    public void updateTimestamp(int i, java.sql.Timestamp t) {
    }
/**

Not implemented

*/

    public void updateTimestamp(java.lang.String s, java.sql.Timestamp t) {
    }
/**

Not implemented

*/

    public void updateString(int i, java.lang.String s) {
    }
/**

Not implemented

*/

    public void updateString(java.lang.String s, java.lang.String t) {
    }

/**

Not implemented

*/
    public int getRow() {
        return 0;
    }

/**

Not implemented

*/
    public java.lang.Object getObject(int i, java.util.Map m) {
        return null;
    }

/**

Not implemented

*/
    public void updateLong(int i, long l) {
    }

/**

Not implemented

*/
    public boolean absolute(int i) {
        return false;
    }

/**

Not implemented

*/
    public void updateNull(int i) {
    }

/**

Not implemented

*/
    public void updateNull(java.lang.String s) {
    }

/**

Not implemented

*/
    public boolean rowUpdated() {
        return false;
    }

/**

Not implemented

*/
    public boolean first() {
        return false;
    }

/**

Not implemented

*/
    public void updateTime(int i, java.sql.Time t) {
    }

/**

Not implemented

*/
    public void updateTime(java.lang.String s, java.sql.Time t) {
    }

/**

Not implemented

*/
    public void afterLast() {
    }

/**

Not implemented

*/
    public void updateAsciiStream(int i, java.io.InputStream s, int j) {
    }

/**

Not implemented

*/
    public void updateAsciiStream(java.lang.String s,
            java.io.InputStream t, int j) {
    }

/**

Not implemented

*/
    public void updateShort(int i, short s) {
    }

/**

Not implemented

*/
    public void updateShort(java.lang.String s, short h) {
    }

/**

Not implemented

*/
    public void updateInt(int i, int j) {
    }

/**

Not implemented

*/
    public void updateInt(java.lang.String s, int i) {
    }

/**

Not implemented

*/
    public void setFetchDirection(int i) {
    }

/**

Not implemented

*/
    public boolean isLast() {
        return false;
    }

/**

Not implemented

*/
    public boolean rowDeleted() {
        return false;
    }

/**

Not implemented

*/
    public boolean isAfterLast() {
        return false;
    }

/**

Not implemented

*/
    public java.sql.Timestamp getTimestamp(int i, java.util.Calendar c) {
        return null;
    }

/**

Not implemented

*/
    public java.sql.Timestamp getTimestamp(java.lang.String s,
            java.util.Calendar c) {
        return null;
    }

/**

Not implemented

*/
    public void updateBinaryStream(int i, java.io.InputStream s, int j) {
    }

/**

Not implemented

*/
    public void updateBinaryStream(java.lang.String s,
            java.io.InputStream i, int j) {
    }

/**

Not implemented

*/
    public void updateObject(int i, java.lang.Object o) {
    }

/**

Not implemented

*/
    public void updateObject(int i, java.lang.Object o, int j) {
    }

/**

Not implemented

*/
    public void updateObject(java.lang.String s, java.lang.Object o) {
    }

/**

Not implemented

*/
    public void updateObject(java.lang.String s, java.lang.Object o,
            int i) {
    }

/**

Not implemented

*/
    public void moveToCurrentRow() {
    }

/**

Not implemented

*/
    public boolean isBeforeFirst() {
        return false;
    }
/**

Not implemented

*/

    public void updateBigDecimal(int i, java.math.BigDecimal b) {
    }
/**

Not implemented

*/

    public void updateBigDecimal(java.lang.String s,
            java.math.BigDecimal b) {
    }
/**

Not implemented

*/

    public void updateBoolean(int i, boolean b) {
    }

/**

Not implemented

*/
    public java.math.BigDecimal getBigDecimal(java.lang.String s) {
        return null;
    }

/**

Not implemented

*/
    public boolean isFirst() {
        return false;
    }

/**

Not implemented

*/
    public java.sql.Statement getStatement() {
        return null;
    }

/**

Not implemented

*/
    public void updateBoolean(java.lang.String s, boolean b) {
    }

/**

Not implemented

*/
    public void insertRow() {
    }

/**

Not implemented

*/
    public boolean previous() {
        return false;
    }
/**

Not implemented

*/

    public void beforeFirst() {
    }
/**

Not implemented

*/

    public void setFetchSize(int i) {
    }

/**

Not implemented

*/
    public java.sql.Array getArray(int i) {
        return null;
    }

/**

Not implemented

*/
    public java.sql.Array getArray(java.lang.String s) {
        return null;
    }

/**

Not implemented

*/
    public java.sql.Blob getBlob(int i) {
        return null;
    }

/**

Not implemented

*/
    public java.sql.Blob getBlob(java.lang.String s) {
        return null;
    }

/**

Not implemented

*/
    public void updateLong(java.lang.String s, long l) {
    }

/**

Not implemented

*/
    public void updateFloat(int i, float f) {
    }

/**

Not implemented

*/
    public void updateFloat(java.lang.String s, float f) {
    }

/**

Not implemented

*/
    public int getType() {
        return 0;
    }

/**

Not implemented

*/
    public boolean last() {
        return false;
    }

/**

Not implemented

*/
    public java.sql.Clob getClob(int i) {
        return null;
    }

/**

Not implemented

*/
    public java.math.BigDecimal getBigDecimal(int i) {
        return null;
    }

/**

Not implemented

*/
    public java.sql.Ref getRef(int i) {
        return null;
    }

/**

Not implemented

*/
    public java.sql.Ref getRef(java.lang.String s) {
        return null;
    }

/**

Not implemented

*/
    public java.sql.Clob getClob(java.lang.String s) {
        return null;
    }
/**

Not implemented

*/

    public java.lang.Object getObject(java.lang.String s, java.util.Map m) {
        return null;
    }

    /**
      handy for testing

      */


    public String toString() {
	StringBuffer buf=new StringBuffer();

try{
            java.sql.ResultSetMetaData r = null;
            java.sql.ResultSet rs = null;
if(resset!=null){
r=resset.getMetaData();
rs=resset;
}else{
r = this.getMetaData();
rs=this;
}
            int count = r.getColumnCount();
            System.out.println("COUNT "+count);

            while (rs.next()) {
		buf.append("ROW: \n");

                for (int k = 0; k < (count); k++) {
                    String key = r.getColumnName(k);
			buf.append("key "+key+"\n");

			SQLNode n=(SQLNode)rs.getObject(key);
			buf.append("value "+n.getContent()+"\n");

                }
            }
        } catch (Exception e) {
            System.err.println("oops "+e);
            e.printStackTrace();
        }


        return buf.toString();
    }


/** 
for RDF resultset. see  
http://www.w3.org/2003/03/rdfqr-tests/recording-query-results.html
http://www.w3.org/2003/03/rdfqr-tests/
*/

    public String toRDFString() {
            java.sql.ResultSetMetaData r = null;
            java.sql.ResultSet rs = null;
	StringBuffer buf2=new StringBuffer();

	StringBuffer buf=new StringBuffer(
"<rdf:RDF "+
"\n    xmlns:rdf='http://www.w3.org/1999/02/22-rdf-syntax-ns#' "+
"\n    xmlns:rs='http://jena.hpl.hp.com/2003/03/result-set#'> "+
"\n   <rs:ResultSet rdf:about=''> ");


        try {

if(resset!=null){
r=resset.getMetaData();
rs=resset;
}else{
r = this.getMetaData();
rs=this;
}


//hm, typing informatioN?
//what's optionsal, and what's not?

Enumeration en=getColumnNames().elements();

while(en.hasMoreElements()){
buf.append("\n\t<rs:resultVariable>"+(String)en.nextElement()+"</rs:resultVariable>");
}


int rows=0;

//            java.sql.ResultSetMetaData r = this.getMetaData();
            int count = r.getColumnCount();
            System.out.println("COUNT "+count);

            while (rs.next()) {
		rows++;

		buf2.append("\n\t<rs:solution>\n\t   <rs:ResultSolution>");

                for (int k = 0; k < (count); k++) {

                    String key = r.getColumnName(k);

			buf2.append("\n\t\t<rs:binding rdf:parseType='Resource'>");
			buf2.append("\n\t\t   <rs:variable>"+key+"</rs:variable>");

//resources??
			SQLNode nn=(SQLNode)rs.getObject(key);

			if(nn.isbnode){
			buf2.append("\n\t\t   <rs:value rdf:nodeID='"+nn.getContent()+"' />");
			}else 
			if(nn.isresource){
			buf2.append("\n\t\t   <rs:value rdf:resource='"+nn.getContent()+"' />");
			}else{
			buf2.append("\n\t\t   <rs:value>"+nn.getContent().trim()+"</rs:value>");
			}
			buf2.append("\n\t\t</rs:binding>");
                }
		buf2.append("\n\t   </rs:ResultSolution>\n\t</rs:solution>");
            }

	buf.append("\n\t<rs:size rdf:datatype='http://www.w3.org/2000/10/XMLSchema#integer'>"+rows+"</rs:size>");
	buf.append(buf2.toString());
	buf.append("\n</rs:ResultSet>");
	buf.append("\n</rdf:RDF>\n");


        } catch (Exception e) {
            System.err.println("oops "+e);
            e.printStackTrace();
        }



        return buf.toString();
    }


/**

Not implemented

*/

public void updateRef(int i,java.sql.Ref ref){

}

/**

Not implemented

*/

public void updateRef(java.lang.String s,java.sql.Ref ref){

}
/**

Not implemented

*/

public void updateBlob(int i,java.sql.Blob b){

}
/**

Not implemented

*/

public void updateBlob(java.lang.String s,java.sql.Blob b){

}
/**

Not implemented

*/

public void updateClob(int i,java.sql.Clob c){

}

/**

Not implemented

*/
public void updateClob(java.lang.String s,java.sql.Clob c){

}

/**

Not implemented

*/
public void updateArray(int i,java.sql.Array a){

}

/**

Not implemented

*/
public void updateArray(java.lang.String s,java.sql.Array a){

}

}

