package org.ilrt.inkling.query;

import org.ilrt.inkling.api.*;
import java.util.*;
import java.security.MessageDigest;
import org.ilrt.inkling.query.squish.*;
import java.io.RandomAccessFile;
import java.security.MessageDigest;
/*
Copyright University of Bristol, 2003

This software is released under the W3C license:
http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231
*/

/**

<pre>
Converts a Squish query to an SQL query, using a two-table model,
triples (subject,predicate,object) [as ints] and
resources(keyhash,value) [int,varchar]

This code is based on work by Matt Biddulph: http://www.hackdiary.com/
</pre>
*/

public class Squish2SQL {

    boolean debug = false;
    static String serverhost = "localhost";


    //bcount is the part which accesses the 'resources' table
    //this list is the list of SQL variable names to access for the resources table
    Vector sqlVariableNamesB = new Vector();

    //index of clauses: number used to name the clauses in sql for triples table
    int clause_idxA = 1;

    //index of clauses: number used to name the clauses in sql for triples table
    int clause_idxB = 1;


    //this is just for the a* variables
    Hashtable realToSqlVariableNameMapA = new Hashtable();

    //table names
    String main_table = "triples";
    String res_table = "resources";

    //strings for building the query
    //new version 2002-11-18 - uses the sid, pid, oid - the internal ids

    String sid = "ssid";//changeme
    String pid = "ppid";
    String oid = "ooid";

    String sss = "subject";
    String ppp = "predicate";
    String ooo = "object";

    //trying to make both versions work, but only the SHA1 one does
    //simple would be just a strings bucket
    boolean SHA1 = true;
    boolean SIMPLE = false;

    //this is for matching the numbers to the clauses between
    //the triples and resources table,
    //there will be an SQL name for the triples table and
    //another for the resource table - we need to keep track
    //of these and force them to refer to the same thing
    //when we build the SQL query.
    //we need a1.subject=b1.subject - that's what this hash does.
    Hashtable sqlVariableMatchAB;


    public Squish2SQL() {
        clause_idxA = 1;
        clause_idxB = 1;
        sqlVariableNamesB = new Vector();
        realToSqlVariableNameMapA = new Hashtable();
        sqlVariableMatchAB = new Hashtable();
    }


    /**

       This method produces a string SQL query from a minimum  of

        * a (full) list of variables  (i.e. not just the ones you want returned (actually
       this could be a problem));

        * the variables you actually want

        * a vector of arrays of clauses: each array: (pred, sub, obj).

        * a vector of arrays of constraints (as pseudo triples at the moment (e.g. array
       is (~ ?y bla)).

@@should be private

       */


    public String triple_sql(Vector allQueryVariables,
            Vector variablesWeWant, Vector clauses, Vector
constraints,boolean prov) {



        //allQueryVariables are all the variables in this query.
        //variableswewant are the variable names that we want to return the value of


        //the string to return
        StringBuffer sql = new StringBuffer("SELECT DISTINCT ");

        try {


            //used for creating 'virtual' SQL tables: e.g. triples as a1, resources as b2
            clause_idxA = 1;

            //we need to keep track fo the real variable names and what we call them
            //in the sql query.
            realToSqlVariableNameMapA = new Hashtable();

            int clause_count = clauses.size();

            //for storing the 'where' parts of the SQL query
            //for the triples and resources tables respectively
            Vector whereA = new Vector();
            Vector whereB = new Vector();

            Enumeration clauseIterator = clauses.elements();



            //---------------------
            //
            // creates a mapping between real and SQl varibales names for A (triples)
            // fills realToSqlVariableNameMapA with these
            //
            //---------------------



            //for each clause
            while (clauseIterator.hasMoreElements()) {


                String[] clause = (String[]) clauseIterator.nextElement();

                //here's our first triple from the list of clauses
                String pred = clause[0];
                String subject = clause[1];
                String object = clause[2];

                int pred_bound = 0;
                int subject_bound = 0;
                int object_bound = 0;


                //this section creates a mapping between the real
                //variables names and their equivalents in the SQL query for the triples table.


                Enumeration queryVariablesIterator =
                        allQueryVariables.elements();

                while (queryVariablesIterator.hasMoreElements()) {

                    String variableNameToMatch =
                            (String) queryVariablesIterator.nextElement();

                    try {

                        //i.e. if variable names are the same in each case
                        if (subject.equals(variableNameToMatch)) {

                            //this adds a mapping from the real name of the variable to its sql name
                            realToSqlVariableNameMapA.put(
                                    variableNameToMatch,
                                    "a"+clause_idxA + "."+sss + "");

                            subject_bound = 1;

                        }

                        if (pred.equals(variableNameToMatch)) {
                            pred_bound = 1;

                            realToSqlVariableNameMapA.put(
                                    variableNameToMatch,
                                    "a"+clause_idxA + "."+ppp + "");

                        }

                        if (object.equals(variableNameToMatch)) {
                            object_bound = 1;

                            realToSqlVariableNameMapA.put(
                                    variableNameToMatch,
                                    "a"+clause_idxA + "."+ooo + "");

                        }

                    } catch (Exception pp) {
                        System.err.println("ERROR "+pp);
                        pp.printStackTrace();
                    }

                }


                //---------------------

                //adding actual clauses to the 'where' part of the query
                //for the triples table e.g. a1.subject='426897987'

                //---------------------

                //sha1 stuff: creating sha1s of s, p, o


                String sh1Sub = "";
                String sh1Pred = "";
                String sh1Obj = "";

                try {


                    sh1Sub = sh1Sub +Util.hashcodeIntFromString(subject);
                    sh1Pred = sh1Pred +Util.hashcodeIntFromString(pred);
                    sh1Obj = sh1Obj +Util.hashcodeIntFromString(object);

                } catch (Exception es) {
                    System.err.println("could not compute sha1 hashcode "+es);
                }


                if (subject_bound == 0) {
                    whereA.addElement("a"+clause_idxA + "."+sss + " = "+
                            sh1Sub + "");
                }

                if (pred_bound == 0) {
                    whereA.addElement("a"+clause_idxA + "."+ppp + " = "+
                            sh1Pred + "");
                }

                if (object_bound == 0) {
                    whereA.addElement("a"+clause_idxA + "."+ooo + " = "+
                            sh1Obj + "");
                }

                clause_idxA++;

            }

            //---------------------
            //
            // creates a vector of SQL variables names for A (triples)
            // fills sqlVariableNamesList with these
            // later we generate paths from this list
            // e.g. a1.subject=a4.object etc
            //
            //---------------------



            Vector sqlVariableNamesList = new Vector();

            int cl_idx = 1;


            //looping through the allQueryVariables again, this time to get the
            //links _between_ the triples variables in the sql query
            //e.g. a1.object=a2.subject and a2.subject=a3.subject
            Enumeration variables = allQueryVariables.elements();
            while (variables.hasMoreElements()) {
                String variableNameToMatch =
                        (String) variables.nextElement();
                Vector sqlVariablesNamesA = new Vector();

                //this is going to be a list of matches from clauses of one variable, and
                //will eventualy be added to bindings

                Enumeration eq = clauses.elements();

                while (eq.hasMoreElements()) {

                    String[] clause = (String[]) eq.nextElement();

                    String pred = clause[0];
                    String subject = clause[1];
                    String object = clause[2];


                    int pred_bound = 0;
                    int subject_bound = 0;
                    int object_bound = 0;


                    if (subject.equals(variableNameToMatch)) {
                        sqlVariablesNamesA.addElement("a"+cl_idx + "."+
                                sss + "");
                    }

                    if (pred.equals(variableNameToMatch)) {
                        sqlVariablesNamesA.addElement("a"+cl_idx + "."+
                                ppp + "");
                    }

                    if (object.equals(variableNameToMatch)) {
                        sqlVariablesNamesA.addElement("a"+cl_idx + "."+
                                ooo + "");
                    }

                    cl_idx++;

                }//end of loop through clauses - still in var loop

                cl_idx = 1;

                if (sqlVariablesNamesA.size() > 1) {
                    sqlVariableNamesList.addElement(sqlVariablesNamesA);
                }


            }//end of loop through var loop



            // adding some more stuff to whereA - these are
            // the sqlVariableNamesList - used to generate links between A-variable names

            for (int i = 0; i < sqlVariableNamesList.size(); i++) {

                Vector bindings =
                        (Vector) sqlVariableNamesList.elementAt(i);

                for (int j = 0; j < bindings.size() - 1; j++) {

                    whereA.addElement((String) bindings.elementAt(j) +
                            "="+(String) bindings.elementAt(j + 1));
                }

            }




            //------------------
            //
            // this part sorts out the mappings from values
            // to the resources table
            // puts these in sqlVariableMatchAB; also adds stuff to the whereB part
            //
            //------------------


            //selectvars is the list of b variables and the actual variables that match them:
            //e.g b1.value as ?x

            Vector selectvars = new Vector();

            //realToSqlVariableNameMapA is a mapping from the real names of the
            //variables: ?x, ?sal etc to the SQL names of the
            //variables: a1.subject, a2.object etc - just the a* (i.e. just the triples table)


            Enumeration qq = realToSqlVariableNameMapA.keys();

            while (qq.hasMoreElements()) {

                String key = (String) qq.nextElement();


                String val = (String) realToSqlVariableNameMapA.get(key);
                String kk = key;
		String vv = val;

                //the keys should not start with "?" any more
                if (key.startsWith("?")) {
                    kk = key.substring(1);
                }


                //the key stays the same but
                //kk should be equal to something in a later clause, and
                //val should be the resource.value, i.e. the value of the key
                //val in the resources table.

                if (variablesWeWant.contains(key)) {


                    //.....literals attempt
                    //change this - put the boolean literal in resources table not triples
                    //leave this for now


                    selectvars.addElement("b"+ clause_idxB + ".value AS "+kk);

//testing - for getting prov back....b? duplicates??
if(prov){
selectvars.addElement("b"+ clause_idxB + ".source AS "+kk+"_id");
}

/*
// for now

selectvars.addElement("b"+ clause_idxB + ".isresource AS "+kk+"_isresource");
selectvars.addElement("b"+ clause_idxB + ".isblank AS "+kk+"_isblank");
selectvars.addElement("b"+ clause_idxB + ".provenance AS "+kk+"_provenance");
*/

                    sqlVariableNamesB.addElement("b"+ clause_idxB);
                    whereB.addElement("b"+ clause_idxB + ".keyhash="+vv);

		String rr=vv.substring(0,vv.indexOf("."));

///whereB.addElement("b"+clause_idxB+".source="+rr+".personid");

                    //need to add in the constraints stuff here
                    //hash of vectors?

                    Vector ttemp;
                    if (!sqlVariableMatchAB.containsKey(val)) {
                        ttemp = new Vector();
                        ttemp.addElement("b"+ clause_idxB + ".value");
                        sqlVariableMatchAB.put(val, ttemp);
                    } else {
                        ttemp = (Vector) sqlVariableMatchAB.get(val);
                        ttemp.addElement("b"+ clause_idxB + ".value");
                    }
                }

                clause_idxB++;
            }





            //------------------
            //
            // building the actual query.
            //
            //
            //------------------



            //real variable names 'x, y' etc

            /*
            	    ///attempt at literal
            	    Enumeration litloop = literalToSqlVariableName.keys();

            	    while(litloop.hasMoreElements()){
            		String key=(String)litloop.nextElement();
            		String val=(String)literalToSqlVariableName.get(key);

                            //the keys should not start with "?" any more
                            if (key.startsWith("?")) {
                                key = key.substring(1);
                            }

            		sql.append(" "+val+" as "+key+", ");
            	    }
            	    ///attempt at literal
            */

            for (int q = 0; q < selectvars.size(); q++) {

                if (q == selectvars.size() - 1) {
                    sql.append(selectvars.elementAt(q).toString() + " ");
                } else {
                    sql.append(selectvars.elementAt(q).toString() + ", ");
                }

            }


            //SQL from clause
            sql.append(" FROM ");

            //from

            for (int z = 1; z < clause_idxA; z++) {
                sql.append(" "+main_table + " a"+z + ", ");
            }


            //variables to get from the resources table

            Enumeration zq = sqlVariableNamesB.elements();

            for (int z = 0; z < sqlVariableNamesB.size(); z++) {

                if (z == (sqlVariableNamesB.size() - 1)) {

                    sql.append(" "+res_table + " "+
                            zq.nextElement().toString() + " ");
                } else {

                    sql.append(" "+res_table + " "+
                            zq.nextElement().toString() + ", ");
                }


            }


            //where for the triples table
            sql.append(" WHERE ");

            for (int r = 0; r < whereA.size(); r++) {

                sql.append(whereA.elementAt(r).toString() + " ");
                sql.append(" AND ");
            }//end for loop

            //where for the resources table

            for (int r = 0; r < whereB.size(); r++) {
		String yy=whereB.elementAt(r).toString();
                sql.append(yy + " ");

                //constraints

                if (constraints == null || constraints.size() == 0) {

                    if (r == whereB.size() - 1) {
                        sql.append(";");
                    } else {
                        sql.append(" AND ");
                    }
                } else {
                    sql.append(" AND ");
                }
            }//end for loop

            if (constraints == null) {

			//System.out.println("no constraints");

		}
            if (constraints != null) {

//System.out.println("got constraints "+constraints);

                for (int zz = 0; zz < constraints.size(); zz++) {


                    String[] tr = (String[]) constraints.elementAt(zz);
                    String key = tr[1].trim();

                    String s = (String) realToSqlVariableNameMapA.get(key);

                    String ss = tr[0];

                    if (SHA1) {

                        //hack. what about multiple values?

                        Vector oo = (Vector) sqlVariableMatchAB.get(s);

                        if (oo != null) {
                            s = oo.elements().nextElement().toString();
                        }
                    }


                    //we need to try to cast this to the right sort of object for the database
                    //(Postgres - don't know about other systems yet)
                    //INT, timestamp, float, date, time
                    //all except numbers must be surrounded by single quotes
                    //should do something clever here with xml datatypes....

                    //** dates not workig yet - needs further thought

                    //ordering? int, float, datetime, date, time?

                    String tocast = tr[2];

                    //shoudl have proper equals for strings - not sure of syntax


                    if (ss.trim().equals("~") ||
                            ss.trim().toLowerCase().equals("like")) {
                        //assume a string
                        //System.out.println("~ or like "+tocast);

                        sql.append("lower("+s + ") like lower('%"+
                                tocast + "%')");
                    } else if (ss.trim().equals("eq")) {

                        sql.append(s + "='"+tocast + "'");

                        //System.out.println("eq "+tocast);
                    } else if (ss.trim().equals("ne")) {

                        //assume a string
                        sql.append(s + "!='"+tocast + "'");
                        //                        sql.append(" not(lower("+s + ") like lower('%"+tocast + "%'))");
                        //System.out.println("neq "+tocast);

                    } else {

                        if (ss.trim().equals("==") ||
                                ss.trim().equals("=")) {
                            ss = "=";
                        }

                        //System.out.println("= "+tocast);

                        try {
                            Integer test = new Integer(tocast);


                            sql.append("CAST(CAST("+s + " as TEXT) as INTEGER) "+
                                    ss + " "+tocast);


                        } catch (Exception e1) {
                            //ErrorLog.write("Not an Integer");

                            try {
                                Float test = new Float(tocast);

                                sql.append("CAST(CAST("+s + " as TEXT) as FLOAT) "+
                                        ss + " "+tocast);

                            } catch (Exception e2) {


                                //ErrorLog.write("Not a Float");


                                //ErrorLog.write("Defaulting to string");

                                //may fail...last resort
                                sql.append(s + " "+ss + " '"+tocast + "'");


                            }//end catch

                        }//end catch

                    }
                    //		    }else{

                    //	               	sql.append(s+" "+ss+" '"+tocast+"'");

                    //                  }//end else



                    if (zz == constraints.size() - 1) {
                        sql.append(";");
                    } else {
                        sql.append(" AND ");
                    }


                }//end for


            }//end if
        }
        catch (Exception yt) {
            yt.printStackTrace();
        }

        if (debug) {
            //System.out.println("SQL query is "+sql.toString());
        }
        return sql.toString();

    }



    public Vector processQueryVariables(Vector r) {

        Vector v = new Vector();

        for (int y = 0; y < r.size(); y++) {

            String s = ((String) r.elementAt(y)).trim();
            if (s.startsWith("?")) {
                s = s.substring(1);
                v.addElement(s);
            }
        }

        return v;

    }


    /**
       basically the main method: takes a parsed query and returns a
       converted SQL query as a string
       */


    public static String convert(String query){
	return convert(query,false);
    }


    /**
       basically the main method: takes a parsed query and returns a
       converted SQL query as a string.
	prov is a boolean represnting whether provenance is retrieved
	or not: if so makes it slower and creates duplicate rows so
	defaults to false
       */

    public static String convert(String query, boolean prov){

        Vector triples= new Vector();
        Vector variables= new Vector();
        Vector variablesWeWant= new Vector();
        Vector consts= new Vector();

            System.out.println(query);

        if(query!=null && (!query.equals(""))){

        ComplexQuery cq=(ComplexQuery)SquishParser.parse(query);

        triples=cq.getTriples();
        variables=cq.getVariables();
        variablesWeWant=cq.getVariablesWeWant();
        consts=cq.getConstraints();
        }

return convert(triples,variables,variablesWeWant,consts,prov);

}

/**

internal conversion method
@@should be private
*/

    public static String convert(Vector triples, Vector variables, Vector
variablesWeWant, Vector constraints, boolean prov) {

        //all the variables
        Vector allQueryVariables = new Vector();

        //just the ones we want
        //Vector variablesWeWant = new Vector();

        //clauses of the query (as arrays)
        Vector clauses = new Vector();

        //constraints of the query (as arrays)
////        Vector constraints = new Vector();//1

        String query = null;
        String query1 = null;

        Enumeration e = triples.elements();

        //we get the clauses as triples - here we convert to arrays
        while (e.hasMoreElements()) {
            Query tt = (Query) e.nextElement();
            String pp = null;
            String ss = null;
            String oo = null;

            if (tt.getPredicate() == null) {
                pp = tt.getPredicateVariable();
            }else{
		pp=tt.getPredicate().getContent();
	    }
            if (tt.getSubject() == null) {
                ss = tt.getSubjectVariable();
            }else{
		ss=tt.getSubject().getContent();
	    }
            if (tt.getObject() == null) {
                oo = tt.getObjectVariable();
            }else{

            	oo = tt.getObject().getContent();

		if(!tt.getObject().isresource){
		oo="\""+oo;
//		System.out.println("testing literals "+oo);
		}
	    }

            String[] clause = {pp, ss, oo};
            clauses.addElement(clause);
        }

///??
        Vector newConstraints = new Vector();

        variablesWeWant = variablesWeWant;
        allQueryVariables = variables;

        Squish2SQL t = new Squish2SQL();

//System.out.println("@@@ "+constraints+ "@@@");

//later....
        //we get the constraints as queries - here we convert to arrays
        Enumeration e1 = constraints.elements();
        while (e1.hasMoreElements()) {

            Query tt =  (Query)e1.nextElement();

		Node ss=tt.getSubject();
		Node pp=tt.getPredicate();
		Node oo=tt.getObject();
		String sss="";
		String ppp="";
		String ooo="";

		if(ss!=null){
		sss=ss.getContent();
		}else{
		sss=tt.getSubjectVariable();
		}

		if(pp!=null){
		ppp=pp.getContent().toString();
//System.out.println("!!!settign "+ppp);
		}else{
		ppp=tt.getPredicateVariable();
		}

		if(oo!=null){
		ooo=oo.getContent();
		}else{
		ooo=tt.getObjectVariable();
		}


            String[] clause = {ppp,sss,ooo};

//tt.getPredicate().getContent(),
//		tt.getSubject().getContent(),
 //           tt.getObject().getContent()};

            newConstraints.addElement(clause);
        }

        String sqlQuery = t.triple_sql(allQueryVariables, variablesWeWant,
                clauses, newConstraints,prov);

        return sqlQuery;

    }

    public static int sha1 (String sha1) throws Exception {
        MessageDigest md = MessageDigest.getInstance("SHA1");
        md.update(sha1.getBytes());
        byte[] digest = md.digest();
        return ((int) digest[0] & 0xff) | (((int) digest[1] & 0xff) << 8) |
                (((int) digest[2] & 0xff) << 16) |
                (((int) digest[3] & 0xff) << 24);
    }


    public static String textFromFile(String fn) throws Exception {
        //System.out.println("file: "+ fn);
        RandomAccessFile in = new RandomAccessFile(fn, "r");
        String text = "";
        String line = "";
        while (line != null) {
            text += line;
            line = in.readLine();
        }
        return text;
    }


            

}
